#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore  #  @manual

from .async_aggregators import (
    FedAdamAsyncAggregatorConfig,
    FedAvgWithLRAsyncAggregatorConfig,
    FedAvgWithLRWithMomentumAsyncAggregatorConfig,
    FedAdamHybridAggregatorConfig,
    FedAvgWithLRHybridAggregatorConfig,
)
from .local_optimizers import LocalOptimizerSGDConfig, LocalOptimizerFedProxConfig
from .optimizer_scheduler import (
    ConstantLRSchedulerConfig,
    LRBatchSizeNormalizerSchedulerConfig,
    ArmijoLineSearchSchedulerConfig,
)
from .server_optimizers import (
    FedAdamOptimizerConfig,
    FedAvgOptimizerConfig,
    FedAvgWithLROptimizerConfig,
    FedLAMBOptimizerConfig,
    FedLARSOptimizerConfig,
)
from .sketch_aggregator import SketchAggregatorConfig
from .sync_aggregators import (
    FedAvgSyncAggregatorConfig,
    FedAdamSyncAggregatorConfig,
    FedAvgWithLRSyncAggregatorConfig,
    FedLARSSyncAggregatorConfig,
    FedLAMBSyncAggregatorConfig,
)

ConfigStore.instance().store(
    name="base_optimizer_sgd",
    node=LocalOptimizerSGDConfig,
    group="optimizer",
)


ConfigStore.instance().store(
    name="base_optimizer_fedprox",
    node=LocalOptimizerFedProxConfig,
    group="optimizer",
)

ConfigStore.instance().store(
    name="base_constant_lr_scheduler",
    node=ConstantLRSchedulerConfig,
    group="lr_scheduler",
)


ConfigStore.instance().store(
    name="base_lr_batch_size_normalizer_scheduler",
    node=LRBatchSizeNormalizerSchedulerConfig,
    group="lr_scheduler",
)


ConfigStore.instance().store(
    name="base_armijo_line_search_lr_scheduer",
    node=ArmijoLineSearchSchedulerConfig,
    group="lr_scheduler",
)


ConfigStore.instance().store(
    name="base_fed_avg_sync_aggregator",
    node=FedAvgSyncAggregatorConfig,
    group="aggregator",
)


ConfigStore.instance().store(
    name="base_fed_avg_with_lr_sync_aggregator",
    node=FedAvgWithLRSyncAggregatorConfig,
    group="aggregator",
)


ConfigStore.instance().store(
    name="base_fed_adam_sync_aggregator",
    node=FedAdamSyncAggregatorConfig,
    group="aggregator",
)


ConfigStore.instance().store(
    name="base_fed_lars_sync_aggregator",
    node=FedLARSSyncAggregatorConfig,
    group="aggregator",
)


ConfigStore.instance().store(
    name="base_fed_lamb_sync_aggregator",
    node=FedLAMBSyncAggregatorConfig,
    group="aggregator",
)


ConfigStore.instance().store(
    name="base_fed_avg_with_lr_async_aggregator",
    node=FedAvgWithLRAsyncAggregatorConfig,
    group="aggregator",
)


ConfigStore.instance().store(
    name="base_fed_avg_with_lr_with_momentum_async_aggregator",
    node=FedAvgWithLRWithMomentumAsyncAggregatorConfig,
    group="aggregator",
)

ConfigStore.instance().store(
    name="base_fed_adam_async_aggregator",
    node=FedAdamAsyncAggregatorConfig,
    group="aggregator",
)


ConfigStore.instance().store(
    name="base_fed_avg_with_lr_hybrid_aggregator",
    node=FedAvgWithLRHybridAggregatorConfig,
    group="aggregator",
)

ConfigStore.instance().store(
    name="base_fed_adam_hybrid_aggregator",
    node=FedAdamHybridAggregatorConfig,
    group="aggregator",
)

ConfigStore.instance().store(
    name="base_sketch_aggregator",
    node=SketchAggregatorConfig,
    group="aggregator",
)

ConfigStore.instance().store(
    name="base_fed_adam",
    node=FedAdamOptimizerConfig,
    group="server_optimizer",
)


ConfigStore.instance().store(
    name="base_fed_avg",
    node=FedAvgOptimizerConfig,
    group="server_optimizer",
)


ConfigStore.instance().store(
    name="base_fed_avg_with_lr",
    node=FedAvgWithLROptimizerConfig,
    group="server_optimizer",
)


ConfigStore.instance().store(
    name="base_fed_lars",
    node=FedLARSOptimizerConfig,
    group="server_optimizer",
)

ConfigStore.instance().store(
    name="base_fed_lamb",
    node=FedLAMBOptimizerConfig,
    group="server_optimizer",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import torch
from torch.optim.optimizer import Optimizer


class LARS(Optimizer):
    r"""Implements LARS algorithm.

    It has been proposed in `Large Batch Training of Convolutional Networks`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        beta (float, optional): coefficient used for computing
            running averages of gradient. (default: 0.9)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
    """

    def __init__(self, params, lr=1e-3, beta=0.9, weight_decay=0):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= beta < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(beta))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = {"lr": lr, "beta": beta, "weight_decay": weight_decay}
        super(LARS, self).__init__(params, defaults)

    @torch.no_grad()
    def get_update(self, p, grad, state, group):

        if group["weight_decay"] != 0:
            grad.add_(p.data, alpha=group["weight_decay"])
        # State initialization
        if len(state) == 0:
            state["step"] = 0
            # Moving averages will be updated _in place_
            # Exponential moving average of gradient values
            state["exp_avg"] = torch.clone(grad).detach()

        # m_{t-1}
        exp_avg = state["exp_avg"]
        beta = group["beta"]

        state["step"] += 1

        # Decay the first moment running average coefficient
        exp_avg.mul_(beta).add_(grad, alpha=1 - beta)

        return exp_avg

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError("LARS does not support sparse gradients")

                state = self.state[p]

                update = self.get_update(p, grad, state, group)
                update_norm = update.pow(2).sum().sqrt()
                weight_norm = p.data.pow(2).sum().sqrt()
                # The LAMB paper suggests bounding the weight norm by some
                # hyperparameters but we choose to eliminate unnecessary
                # hyperparameters
                scaling_function = weight_norm
                assert update_norm != 0

                update.mul_(scaling_function / update_norm)
                p.data.add_(update, alpha=-group["lr"])

        return loss


class LAMB(LARS):
    r"""Implements LAMB algorithm.

    It has been proposed in `Large Batch Optimization for Deep Learning: Training BERT in 76 minutes`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        beta1 (float, optional): coefficient used for computing
            running averages of gradient (default 0.9)
        beta2 (float, optional): coefficient used for computing
            running average of gradient squared (default 0.999)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)


    """

    def __init__(
        self, params, lr=1e-3, beta1=0.9, beta2=0.999, eps=1e-8, weight_decay=0
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= beta1 < 1.0:
            raise ValueError("Invalid beta1: {}".format(beta1))
        if not 0.0 <= beta2 < 1.0:
            raise ValueError("Invalid beta2: {}".format(beta2))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = {
            "lr": lr,
            "beta1": beta1,
            "beta2": beta2,
            "eps": eps,
            "weight_decay": weight_decay,
        }
        Optimizer.__init__(self, params, defaults)

    @torch.no_grad()
    def get_update(self, p, grad, state, group):
        # State initialization
        if len(state) == 0:
            state["step"] = 0

            # Moving averages will be updated _in place_
            # Exponential moving average of gradient values
            state["exp_avg"] = torch.zeros_like(p, memory_format=torch.preserve_format)
            # Exponential moving average of squared gradient values
            state["exp_avg_sq"] = torch.zeros_like(
                p, memory_format=torch.preserve_format
            )

        # m_{t-1} and v_{t-1}
        exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
        beta1 = group["beta1"]
        beta2 = group["beta2"]

        state["step"] += 1
        bias_correction1 = 1 - beta1 ** state["step"]
        bias_correction2 = 1 - beta2 ** state["step"]

        # m_t = (beta1 * m_{t-1} + (1-beta1)*g_t)
        exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

        # v_t = (beta2 * v_{t-1} + (1-beta2)*g_t^2)
        exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

        m_update = torch.div(exp_avg, bias_correction1)
        v_update = torch.div(exp_avg_sq, bias_correction2)

        # denom = sqrt(v_t) + eps
        denom = torch.add(v_update.sqrt(), group["eps"])

        # update = l2-penalty + m_{t} / denom
        m_update.div_(denom)
        if group["weight_decay"] != 0:
            m_update.add_(p.data, alpha=group["weight_decay"])

        return m_update



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore  #  @manual

from .sync_dp_servers import SyncDPSGDServerConfig
from .sync_secagg_servers import SyncSecAggServerConfig
from .sync_servers import SyncServerConfig

ConfigStore.instance().store(
    name="base_sync_server",
    node=SyncServerConfig,
    group="server",
)

ConfigStore.instance().store(
    name="base_sync_dp_server",
    node=SyncDPSGDServerConfig,
    group="server",
)

ConfigStore.instance().store(
    name="base_sync_secagg_server",
    node=SyncSecAggServerConfig,
    group="server",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""
This file defines the concept of a differentially private
client where a sample level dp is enforced during training.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

from flsim.channels.base_channel import IdentityChannel
from flsim.clients.base_client import Client, ClientConfig
from flsim.common.timeout_simulator import TimeOutSimulator
from flsim.data.data_provider import IFLUserData
from flsim.interfaces.model import IFLModel
from flsim.privacy.common import PrivacyBudget, PrivacySetting
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.cuda import ICudaStateManager, DEFAULT_CUDA_MANAGER
from opacus import PrivacyEngine


class DPClient(Client):
    def __init__(
        self,
        *,
        dataset: IFLUserData,
        channel: Optional[IdentityChannel] = None,
        timeout_simulator: Optional[TimeOutSimulator] = None,
        store_last_updated_model: Optional[bool] = False,
        name: Optional[str] = None,
        cuda_manager: ICudaStateManager = DEFAULT_CUDA_MANAGER,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=DPClientConfig,
            **kwargs,
        )

        super().__init__(
            dataset=dataset,
            channel=channel,
            timeout_simulator=timeout_simulator,
            store_last_updated_model=store_last_updated_model,
            name=name,
            cuda_manager=cuda_manager,
            **kwargs,
        )
        self.dataset_length = -1
        self.privacy_steps = 0
        self._privacy_budget = PrivacyBudget()
        self.privacy_on = (
            # pyre-fixme[16]: `DPClient` has no attribute `cfg`.
            self.cfg.privacy_setting.noise_multiplier >= 0
            and self.cfg.privacy_setting.clipping_value < float("inf")
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _get_dataset_stats(self, model: IFLModel):
        for batch in self.dataset:
            batch_size = model.get_num_examples(batch)
            break
        # pyre-fixme[61]: `batch_size` may not be initialized here.
        return batch_size, self.dataset.num_examples()

    @property
    def privacy_budget(self) -> PrivacyBudget:
        return self._privacy_budget

    def prepare_for_training(self, model: IFLModel):
        """
        1- call parent's prepare_for_training
        2- attach the privacy_engine
        """
        model, optimizer, optimizer_scheduler = super().prepare_for_training(model)
        if self.privacy_on:
            batch_size, self.dataset_length = self._get_dataset_stats(model)
            privacy_engine = PrivacyEngine(
                module=model.fl_get_module(),
                batch_size=batch_size,
                sample_size=self.dataset_length,
                # pyre-fixme[16]: `DPClient` has no attribute `cfg`.
                alphas=self.cfg.privacy_setting.alphas,
                noise_multiplier=self.cfg.privacy_setting.noise_multiplier,
                max_grad_norm=self.cfg.privacy_setting.clipping_value,
                delta=self.cfg.privacy_setting.target_delta,
            )
            if self.cfg.privacy_setting.noise_seed is not None:
                privacy_engine.random_number_generator = privacy_engine._set_seed(
                    self.cfg.privacy_setting.noise_seed
                )
            privacy_engine.steps = self.privacy_steps
            privacy_engine.attach(optimizer)
        return model, optimizer, optimizer_scheduler

    def _get_privacy_budget(self, optimizer) -> PrivacyBudget:
        if self.privacy_on and self.dataset_length > 0:
            eps, delta = optimizer.privacy_engine.get_privacy_spent()
            return PrivacyBudget(epsilon=eps, delta=delta)
        else:
            return PrivacyBudget()

    def post_batch_train(
        self, epoch: int, model: IFLModel, sample_count: int, optimizer: Any
    ):
        if self.privacy_on and sample_count > optimizer.privacy_engine.batch_size:
            raise ValueError(
                "Batchsize was not properly calculated!"
                " Calculated Epsilons are not Correct"
            )

    def post_train(self, model: IFLModel, total_samples: int, optimizer: Any):
        if not self.privacy_on:
            self.logger.debug(f"Privacy Engine is not enabled for client: {self.name}!")
            return
        if self.dataset_length != total_samples:
            DPClient.logger.warning(
                "Calculated privacy budgets were not Accurate." " Fixing the problem."
            )
            sample_rate = float(optimizer.privacy_engine.batch_size) / total_samples
            optimizer.privacy_engine.sample_rate = sample_rate

        self._privacy_budget = self._get_privacy_budget(optimizer)
        DPClient.logger.debug(f"Privacy Budget: {self._privacy_budget}")

        # book keeping
        privacy_engine = optimizer.privacy_engine
        self.privacy_steps = privacy_engine.steps
        # detach the engine to be safe, (not necessary if model is not reused.)
        privacy_engine.detach()
        # re-add the detached engine so that can be saved along with optimizer
        optimizer.privacy_engine = privacy_engine


@dataclass
class DPClientConfig(ClientConfig):
    """
    Contains configurations for a dp user (sample-level dp)
    """

    _target_: str = fullclassname(DPClient)
    privacy_setting: PrivacySetting = PrivacySetting()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import logging
from queue import PriorityQueue
from typing import Optional

from flsim.channels.base_channel import IdentityChannel
from flsim.clients.async_client import AsyncClientDevice, AsyncClientFactory
from flsim.clients.base_client import ClientConfig
from flsim.common.logger import Logger
from flsim.common.timeout_simulator import TimeOutSimulator
from flsim.common.training_event_handler import AsyncTrainingEventHandler
from flsim.utils.async_trainer.async_user_selector import AsyncUserSelector
from flsim.utils.async_trainer.training_event_generator import (
    IEventGenerator,
)
from flsim.utils.cuda import ICudaStateManager, DEFAULT_CUDA_MANAGER
from tqdm import tqdm


class JobQueueStats:
    """Keeps track of #pending jobs in a job queue.
    Computes event based average of jobs
    """

    def __init__(self):
        self.num_finished_jobs: float = 0
        self.pending_jobs_sum: float = 0
        self.num_started_jobs: float = 0

    def on_job_start(self) -> None:
        self.num_started_jobs += 1

    def on_job_end(self) -> None:
        self.num_finished_jobs += 1
        self.pending_jobs_sum += self.num_started_jobs
        self.num_started_jobs -= 1

    def avg_pending_jobs(self) -> float:
        if not self.num_finished_jobs:
            return float("Inf")
        return self.pending_jobs_sum / self.num_finished_jobs

    def as_str(self) -> str:
        return f"Mean:{self.avg_pending_jobs():.3f}"


class AsyncTrainingSimulator:
    """This class simulates asynchronous training of devices where training
    start times are assumed to be drawn from some probability Generator
    (e.g. IEventGenerator). It has a priority queue to keep track
    of event starts and event ends. The queue will have exactly one start
    event at any point (start events are typically poisson distributed).
    Exactly how many end events are there depends on the Generator of
    training time compared to the rate of the poisson process for event
    starts. It is initialized with:
    1) event_generator: IEventGenerator,
    2) model_trainer: AsyncTrainingEventHandler, and
    3) num_train_end_events_per_epoch: int.
    Async trainer's top-level run() function will be called for each epoch.
    Then, the simulator will first fetch a DeviceState from its min_heap and
    updates the simulator’s timestamp to the event’s relevant_time. If the
    fetched DeviceState is waiting for training, we start training with
    AsyncJobScheduler’s on_training_start_function, put the DeviceState back
    to the min_heap, and log to the JobQueueStats.
    Otherwise, it simply ends the training. Note that this works, because
    relevant_tmie is set to training end time after training is triggered.
    """

    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(
        self,
        event_generator: IEventGenerator,
        job_scheduler: AsyncTrainingEventHandler,
        user_selector: AsyncUserSelector,
        shared_client_config: ClientConfig,
        num_train_end_events_per_epoch: int,
        cuda_manager: ICudaStateManager = DEFAULT_CUDA_MANAGER,
        timeout_simulator: Optional[TimeOutSimulator] = None,
        channel: Optional[IdentityChannel] = None,
    ):
        self.job_scheduler: AsyncTrainingEventHandler = job_scheduler
        self.user_selector: AsyncUserSelector = user_selector
        self.event_generator: IEventGenerator = event_generator
        self.shared_client_config = shared_client_config
        self.num_train_end_events_per_epoch: int = num_train_end_events_per_epoch
        self.num_train_end_events: int = 0
        self.current_time: float = 0
        self.min_heap: PriorityQueue[AsyncClientDevice] = PriorityQueue()
        self.queue_stats: JobQueueStats = JobQueueStats()
        self.timeout_simulator = timeout_simulator
        self.channel = channel
        self.cuda_manager = cuda_manager
        # init the first event
        self.create_future_training_start_event()

    def create_future_training_start_event(self) -> None:
        # create training client and insert into heap
        new_client = AsyncClientFactory.create(
            current_time=self.current_time,
            event_generator=self.event_generator,
            user_selector=self.user_selector,
            client_config=self.shared_client_config,
            timeout_simulator=self.timeout_simulator,
            channel=self.channel,
            cuda_manager=self.cuda_manager,
        )
        # put will query __lt__ on top
        self.min_heap.put(new_client)

    def start_training(self, top: AsyncClientDevice) -> None:
        self.job_scheduler.on_training_start(top)
        # put will query __lt__ on top
        self.min_heap.put(top)
        self.queue_stats.on_job_start()

    def end_training(self, top: AsyncClientDevice) -> None:
        self.queue_stats.on_job_end()
        self.job_scheduler.on_training_end(top)

    def run_one_epoch(self) -> None:
        with tqdm(
            total=self.num_train_end_events_per_epoch,
            desc="Client Training Per Epoch",
            unit="client",
            position=0,
        ) as pbar:
            while not self.min_heap.empty():
                # notify async trainer that num_train_end_events_per_epoch have ended training
                if self.num_train_end_events == self.num_train_end_events_per_epoch:
                    self.num_train_end_events = 0
                    break

                top = self.min_heap.get()
                # forward time
                self.current_time = top.next_event_time()

                if top.is_waiting_to_start():
                    self.start_training(top)
                    self.create_future_training_start_event()
                else:
                    self.num_train_end_events += 1
                    self.end_training(top)
                    pbar.update(1)

    def avg_pending_jobs(self) -> float:
        return self.queue_stats.avg_pending_jobs()

    def print_stats(self, prefix: str) -> None:
        print(f"{prefix}PendingJobs, {self.queue_stats.as_str()}")
        print(f"Current Time: {self.current_time:.4f}")



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore  #  @manual

from .base_round_reducer import RoundReducerConfig
from .dp_round_reducer import DPRoundReducerConfig
from .secure_round_reducer import SecureRoundReducerConfig
from .sketch_round_reducer import SketchRoundReducerConfig
from .weighted_dp_round_reducer import WeightedDPRoundReducerConfig


ConfigStore.instance().store(
    name="base_reducer",
    node=RoundReducerConfig,
    group="reducer",
)

ConfigStore.instance().store(
    name="base_dp_reducer",
    node=DPRoundReducerConfig,
    group="reducer",
)

ConfigStore.instance().store(
    name="base_secure_reducer",
    node=SecureRoundReducerConfig,
    group="reducer",
)

ConfigStore.instance().store(
    name="base_weighted_dp_reducer",
    node=WeightedDPRoundReducerConfig,
    group="reducer",
)

ConfigStore.instance().store(
    name="base_sketch_reducer",
    node=SketchRoundReducerConfig,
    group="reducer",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
from typing import Any, Dict, List, NamedTuple, Optional, Union, Tuple
from unittest.mock import MagicMock

import numpy as np
import torch
import torch.nn as nn
from flsim.common.logger import Logger
from flsim.data.data_provider import IFLUserData
from flsim.interfaces.batch_metrics import IFLBatchMetrics
from flsim.interfaces.metrics_reporter import (
    Channel,
    IFLMetricsReporter,
    Metric,
    TrainingStage,
)
from flsim.interfaces.model import IFLModel
from flsim.metrics_reporter.tensorboard_metrics_reporter import FLMetricsReporter
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.simple_batch_metrics import FLBatchMetrics
from torch.utils.tensorboard import SummaryWriter


class DatasetFromList:
    """
    Simple dataset from  a list of tuples.
    Each item in the outer list will be a batch and each
    batch itself is a tuple of two lists, the raw_batch
    and the batch.
    """

    def __init__(self, list_dataset):
        self.ds = list_dataset

    def __len__(self):
        _, batch = self.ds[0]
        return len(self.ds) * len(batch)

    def __iter__(self):
        return iter(self.ds)


class DummyUserData(IFLUserData):
    def __init__(self, data, model, from_data_provider=False):
        self.data = data
        self._num_examples: int = 0
        self._num_batches: int = 0
        self.model = model
        self.from_data_provider = from_data_provider
        for _, batch in self.data:
            self._num_examples += (
                batch["label"].shape[0] if self.from_data_provider else batch.shape[0]
            )
            self._num_batches += 1

    def __iter__(self):
        for _, batch in self.data:
            yield self.model.fl_create_training_batch(batch=batch)

    def num_batches(self):
        return self._num_batches

    def num_examples(self):
        return self._num_examples


class Quadratic1D(nn.Module):
    """
        a toy optimization example:
            min f(x) = 100 x^2 - 1

    minima is x=0.0, x is initialized at 1.0.
    """

    def __init__(self):
        super(Quadratic1D, self).__init__()
        self.x = nn.Parameter(torch.ones(1))
        self.y = torch.tensor([1.0])

    def forward(self):
        return 100 * torch.square(self.x) - self.y


class MockQuadratic1DFL(IFLModel):
    """
    a dummy IFL wrapper for Quadratic1D
    """

    def __init__(self, model):
        self.model = model

    def fl_forward(self, data=None):
        loss = self.model()
        return FLBatchMetrics(
            loss=loss, num_examples=1, predictions=None, targets=None, model_inputs=None
        )

    def fl_create_training_batch(self):
        pass

    def fl_cuda(self):
        pass

    def fl_get_module(self):
        return self.model

    def get_eval_metrics(self):
        pass

    def get_num_examples(self):
        pass


class TwoFC(nn.Module):
    def __init__(self):
        super().__init__()
        self.fc1 = nn.Linear(2, 5)
        self.fc2 = nn.Linear(5, 1)

    def forward(self, x):
        x = self.fc1(x)
        x = self.fc2(x)
        return x

    def fill_all(self, value):
        def fill(layer):
            if type(layer) == nn.Linear:
                layer.bias.data.fill_(value)
                layer.weight.data.fill_(value)

        self.apply(fill)


class Metrics(IFLBatchMetrics):
    def __init__(self, num_examples, loss):
        self._num_examples = num_examples
        self._loss = loss

    @property
    def loss(self) -> torch.Tensor:
        return self._loss

    @property
    def num_examples(self) -> int:
        return self._num_examples

    @property
    def predictions(self):
        pass

    @property
    def targets(self):
        pass

    @property
    def model_inputs(self):
        pass


class FakeMetricReporter(IFLMetricsReporter):
    def add_batch_metrics(self, metrics: IFLBatchMetrics) -> None:
        pass

    def aggregate(self, one_user_metrics: IFLMetricsReporter):
        pass

    def report_metrics(
        self,
        reset: bool,
        stage: TrainingStage,
        extra_metrics: Optional[List[Metric]] = None,
        **kwargs,
    ) -> Tuple[Any, bool]:
        return (None, False)

    def reset(self):
        pass

    def compare_metrics(self, eval_metrics, best_metrics):
        pass


class SimpleMetricReporter(FakeMetricReporter):
    def __init__(self):
        self.batch_metrics = []

    def add_batch_metrics(self, metrics: IFLBatchMetrics) -> None:
        self.batch_metrics.append(metrics)


class SampleNet(IFLModel):
    def __init__(self, model):
        self.sample_nn = model
        self._num_examples = None
        self._out = None

    def fl_forward(self, batch):
        y = self.sample_nn(batch)
        return Metrics(len(batch), y.mean())

    def fl_create_training_batch(self, batch=None, **kwargs):
        return batch

    def fl_get_module(self):
        return self.sample_nn

    def fl_cuda(self):
        pass

    def get_eval_metrics(self, batch):
        pass

    def get_num_examples(self, batch):
        return len(batch)


class SampleNetHive(SampleNet):
    def __init__(self, value=None):
        self.sample_nn = TwoFC()
        if value is not None:
            self.sample_nn.fill_all(value)
        self._num_examples = None
        self._out = None

    def fl_forward(self, batch):
        x, y = batch["user_n"], batch["label"]
        x = x.flatten().repeat(2).float()
        y = y.flatten().float()
        preds = self.sample_nn(x)
        loss = nn.BCEWithLogitsLoss()(preds, y)
        return Metrics(self.get_num_examples(batch), loss=loss)

    def get_num_examples(self, batch):
        return len(batch["label"])


def verify_models_equivalent_after_training(
    model1: Union[nn.Module, IFLModel],
    model2: Union[nn.Module, IFLModel],
    model_init: Optional[Union[nn.Module, IFLModel]] = None,
    rel_epsilon: Optional[float] = None,
    abs_epsilon: Optional[float] = None,
) -> str:
    """This function accepts either nn.Module or IFLModel and checks that:
    a) Model training did something:
        model1 & model2 are different from model_init
    b) model1 and model2 have same parameters

    Return value: str. "" if both a) and b) are satisfied.
        else, error message with the SAD (sum of absolute difference between
        mismatched model parameters)
    """
    model1 = model1.fl_get_module() if isinstance(model1, IFLModel) else model1
    model2 = model2.fl_get_module() if isinstance(model2, IFLModel) else model2
    if model_init is not None:
        model_init = (
            model_init.fl_get_module()
            if isinstance(model_init, IFLModel)
            else model_init
        )
        # Ensure that training actually did something to model 1.
        if (
            FLModelParamUtils.get_mismatched_param(
                [model1, model_init], rel_epsilon=rel_epsilon, abs_epsilon=abs_epsilon
            )
            == ""
        ):
            return "Model 1 training did nothing"

        # Ensure that training actually did something to model 2.
        if (
            FLModelParamUtils.get_mismatched_param(
                [model2, model_init], rel_epsilon=rel_epsilon, abs_epsilon=abs_epsilon
            )
            == ""
        ):
            return "Model 2 training did nothing"

    # check models identical under both configs
    mismatched_param = FLModelParamUtils.get_mismatched_param(
        [model1, model2], rel_epsilon=rel_epsilon, abs_epsilon=abs_epsilon
    )

    if mismatched_param != "":
        summed_absolute_difference = (
            (
                model1.state_dict()[mismatched_param]
                - model2.state_dict()[mismatched_param]
            )
            .abs()
            .sum()
        )
        return (
            f"Model 1, Model 2 mismatch. Param: {mismatched_param},"
            f" Summed Absolute Difference={summed_absolute_difference}"
        )
    else:
        return ""


def model_parameters_equal_to_value(model, value):
    if isinstance(model, IFLModel):
        model = model.fl_get_module()
    for n, p in model.named_parameters():
        if not torch.allclose(p.float(), torch.tensor(value)):
            summed_absolute_difference = (p - torch.tensor(value)).abs().sum()
            return (
                n
                + f"{p} did not match with {value}: Summed Absolute Difference={summed_absolute_difference}"
            )
    return ""


def check_inherit_logging_level(obj: Any, level: int) -> bool:
    Logger.set_logging_level(level)
    return obj.logger.getEffectiveLevel() == level


class MockRecord(NamedTuple):
    tag: str = ""
    value: Union[float, Dict[str, float]] = 0.0
    global_step: int = 0
    walltime: float = 0


class MetricsReporterWithMockedChannels(FLMetricsReporter):
    """Simulates an FL reporter with STDOUT and Tensorboard channels
    STDOUT and Tensorboard channels are mocked
    """

    def __init__(self):
        super().__init__([Channel.STDOUT, Channel.TENSORBOARD])

        self.tensorboard_results: List[MockRecord] = []
        self.stdout_results: List[MockRecord] = []

        def add_scalar(tag, scalar_value, global_step=None, walltime=None):
            self.tensorboard_results.append(
                MockRecord(tag, scalar_value, global_step, walltime)
            )
            return self

        def add_scalars(main_tag, tag_scalar_dict, global_step=None, walltime=None):
            for tag, value in tag_scalar_dict.items():
                self.tensorboard_results.append(
                    MockRecord(f"{main_tag}/{tag}", value, global_step, walltime)
                )

        def printer(*args):
            self.stdout_results.append(tuple(arg for arg in args))

        SummaryWriter.add_scalar = MagicMock(side_effect=add_scalar)
        SummaryWriter.add_scalars = MagicMock(side_effect=add_scalars)
        self.print = MagicMock(side_effect=printer)

    def compare_metrics(self, eval_metrics, best_metrics) -> bool:
        return True

    def compute_scores(self) -> Dict[str, Any]:
        return {}

    def create_eval_metrics(
        self, scores: Dict[str, Any], total_loss: float, **kwargs
    ) -> Any:
        return None

    def set_summary_writer(self, log_dir: Optional[str]):
        super().set_summary_writer("/tmp/")


class RandomEvalMetricsReporter(IFLMetricsReporter):
    """This metrics reporter is useful for unit testing. It does four things:
    a) When report_metrics(stage=Eval) is called, it produced a random
       eval result
    b) It keeps track of the best eval result produced, and the model that
       produced this result.
    c) When report_metrics(stage=Test) is called, it returns the best eval result
       Why? This is useful in testing the trainer.train() function, which returns
       test eval results.
    d) It can be queried to return the best_model, and the value of
       the best result

    """

    def __init__(self):
        self._best_eval_result: float = -1
        self._best_eval_model: IFLModel = None

    def add_batch_metrics(self, metrics: IFLBatchMetrics) -> None:
        pass

    def aggregate(self, one_user_metrics: "IFLMetricsReporter"):
        pass

    def report_metrics(
        self,
        reset: bool,
        stage: TrainingStage,
        extra_metrics: Optional[List[Metric]] = None,
        **kwargs,
    ) -> Tuple[Any, bool]:
        if stage != TrainingStage.EVAL:
            return self._best_eval_result, False

        assert "model" in kwargs.keys(), f"Did not find model in kwargs: {kwargs}"
        model: IFLModel = kwargs.get("model", None)
        eval_result = np.random.random_sample()

        if eval_result > self._best_eval_result:
            print(
                f"MetricReporter current_eval:{eval_result}, best_eval: {self._best_eval_result}"
            )
            self._best_eval_model = copy.deepcopy(model)
            self._best_eval_result = eval_result
            return (eval_result, True)
        else:
            return (eval_result, False)

    def reset(self):
        pass

    @property
    def best_eval_result(self) -> float:
        return self._best_eval_result

    @property
    def best_eval_model(self) -> IFLModel:
        return self._best_eval_model


def create_model_with_value(value) -> nn.Module:
    model = TwoFC()
    model.fill_all(value)
    return model



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterable, List, Optional, Tuple

import torch
from flsim.clients.base_client import Client
from flsim.clients.dp_client import DPClientConfig, DPClient
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.metrics_reporter import IFLMetricsReporter, Metric
from flsim.interfaces.model import IFLModel
from flsim.reducers.dp_round_reducer import DPRoundReducerConfig
from flsim.trainers.sync_trainer import SyncTrainer, SyncTrainerConfig, Timeline
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from hydra.utils import instantiate
from omegaconf import OmegaConf


class PrivateSyncTrainer(SyncTrainer):
    r"""
    A ``SyncTrainer`` that supports both sample level
    and user level differential privacy (DP).
    """

    def __init__(
        self,
        *,
        model: IFLModel,
        cuda_enabled: bool = False,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=PrivateSyncTrainerConfig,
            **kwargs,
        )
        raise ValueError(
            "PrivateSyncTrainer is deprecated. Please use SyncTrainer with SyncDPServerConfig for user-level dp or with DPClientConfig for sample-level DP"
        )
        super().__init__(model=model, cuda_enabled=cuda_enabled, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def train(
        self,
        data_provider: IFLDataProvider,
        metric_reporter: IFLMetricsReporter,
        num_total_users: int,
        distributed_world_size: int,
        rank: int = 0,
    ) -> Tuple[IFLModel, Any]:
        dp_round_reducer = instantiate(
            # pyre-fixme[16]: `PrivateSyncTrainer` has no attribute `cfg`.
            self.cfg.reducer,
            global_model=self.global_model(),
            num_users_per_round=self.cfg.users_per_round,
            total_number_of_users=num_total_users,
            channel=self.channel,
        )
        # pyre-ignore[16]
        self.aggregator.init_round(dp_round_reducer)
        return super().train(
            data_provider,
            metric_reporter,
            num_total_users,
            distributed_world_size,
            rank,
        )

    def create_or_get_client_for_data(self, dataset_id: int, datasets: Any):
        """This function is used to create clients in a round. Thus, it
        is called UPR * num_rounds times per training run. Here, we use
        <code>OmegaConf.structured</code> instead of <code>hydra.instantiate</code>
        to minimize the overhead of hydra object creation.
        """
        self.clients[dataset_id] = DPClient(
            # pyre-ignore [16]: `PrivateSyncTrainer` has no attribute `cfg`
            **OmegaConf.structured(self.cfg.client),
            dataset=datasets[dataset_id],
            name=f"client_{dataset_id}",
            timeout_simulator=self._timeout_simulator,
            store_last_updated_model=self.cfg.report_client_metrics,
            channel=self.channel,
        )
        return self.clients[dataset_id]

    def calc_post_aggregation_train_metrics(
        self,
        clients: Iterable[Client],
        model: IFLModel,
        timeline: Timeline,
        metric_reporter: Optional[IFLMetricsReporter],
    ) -> List[Metric]:

        metrics = super().calc_post_aggregation_train_metrics(
            clients, model, timeline, metric_reporter
        )

        # calculate sample level dp privacy loss statistics.
        all_client_eps = torch.Tensor(
            [c.privacy_budget.epsilon for c in clients]  # pyre-ignore
        )
        mean_client_eps = all_client_eps.mean()
        max_client_eps = all_client_eps.max()
        min_client_eps = all_client_eps.min()
        p50_client_eps = torch.median(all_client_eps)
        sample_dp_metrics = Metric.from_args(
            mean=mean_client_eps,
            min=min_client_eps,
            max=max_client_eps,
            median=p50_client_eps,
        )

        # calculate user level dp privacy loss statistics.
        # pyre-ignore[16]
        dp_round_reducer = self.aggregator.reducer
        aggr_eps = dp_round_reducer.privacy_budget.epsilon

        return metrics + [
            Metric("sample level dp (eps)", sample_dp_metrics),
            Metric("user level dp (eps)", aggr_eps),
        ]


@dataclass
class PrivateSyncTrainerConfig(SyncTrainerConfig):
    _target_: str = fullclassname(PrivateSyncTrainer)
    client: DPClientConfig = DPClientConfig()  # For sample-level DP
    # To disable sample-level DP, set DPClientConfig.clipping_value = float("inf")
    reducer: DPRoundReducerConfig = DPRoundReducerConfig()  # For user-level DP
    # To disable user-level DP, set DPRoundReducerConfig.clipping_value = float("inf")



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
from dataclasses import dataclass
from typing import List

import numpy as np
import pytest
from flsim.active_user_selectors.simple_user_selector import (
    UniformlyRandomActiveUserSelectorConfig,
)
from flsim.channels.base_channel import IdentityChannel
from flsim.channels.half_precision_channel import HalfPrecisionChannel
from flsim.channels.message import Message
from flsim.channels.sparse_mask_channel import SparseMaskChannelConfig
from flsim.common.pytest_helper import assertEqual, assertEmpty
from flsim.common.pytest_helper import assertRaises
from flsim.optimizers.server_optimizers import (
    FedAvgOptimizerConfig,
    FedAdamOptimizerConfig,
    FedAvgWithLROptimizerConfig,
    FedLARSOptimizerConfig,
    FedLAMBOptimizerConfig,
    OptimizerType,
)
from flsim.reducers.sketch_round_reducer import (
    SketchRoundReducerConfig,
)
from flsim.servers.aggregator import AggregationType
from flsim.servers.sync_servers import SyncServerConfig
from flsim.tests.utils import (
    create_model_with_value,
    model_parameters_equal_to_value,
    verify_models_equivalent_after_training,
    SampleNet,
    TwoFC,
)
from flsim.utils.fl.common import FLModelParamUtils
from hydra.errors import HydraException
from hydra.utils import instantiate


@dataclass
class MockClientUpdate:
    deltas: List[float]
    weights: List[float]
    expected_value: float


class TestSyncServer:
    def _create_client_updates(self, num_clients, aggregation_type):
        deltas = [i + 1 for i in range(num_clients)]
        weights = [i + 1 for i in range(num_clients)]

        if aggregation_type == AggregationType.WEIGHTED_AVERAGE:
            expected_value = float(np.average(deltas, weights=weights))
        elif aggregation_type == AggregationType.AVERAGE:
            expected_value = float(np.average(deltas))
        elif aggregation_type == AggregationType.WEIGHTED_SUM:
            expected_value = float(sum(d * w for d, w in zip(deltas, weights)))
        elif aggregation_type == AggregationType.SUM:
            expected_value = float(sum(deltas))
        return MockClientUpdate(deltas, weights, expected_value)

    def _run_one_round_comparison(
        self,
        optimizer,
        server,
        optim_model,
        server_model,
        client_updates,
    ):
        server.init_round()
        optimizer.zero_grad()
        for delta, weight in zip(client_updates.deltas, client_updates.weights):
            server.receive_update_from_client(
                Message(model=SampleNet(create_model_with_value(delta)), weight=weight)
            )

        FLModelParamUtils.set_gradient(
            model=optim_model,
            reference_gradient=create_model_with_value(client_updates.expected_value),
        )
        optimizer.step()
        server.step()
        return server_model, optim_model

    def _compare_optim_and_server(self, opt_config, num_rounds, num_clients, agg_type):
        server_model = SampleNet(create_model_with_value(0))
        optim_model = create_model_with_value(0)
        server = instantiate(
            SyncServerConfig(aggregation_type=agg_type, server_optimizer=opt_config),
            global_model=server_model,
        )

        optimizer = OptimizerType.create_optimizer(optim_model, opt_config)
        client_updates = self._create_client_updates(
            num_clients, aggregation_type=agg_type
        )

        for _ in range(num_rounds):
            server_model, optim_model = self._run_one_round_comparison(
                optimizer, server, optim_model, server_model, client_updates
            )
            error_msg = verify_models_equivalent_after_training(
                server_model, optim_model
            )
            assertEmpty(error_msg, msg=error_msg)

    @pytest.mark.parametrize(
        "aggregation_type",
        [
            AggregationType.AVERAGE,
            AggregationType.WEIGHTED_AVERAGE,
        ],
    )
    @pytest.mark.parametrize(
        "num_clients",
        [10, 1],
    )
    @pytest.mark.parametrize(
        "num_rounds",
        [10, 1],
    )
    def test_fed_avg_sync_server(
        self, aggregation_type, num_clients, num_rounds
    ) -> None:
        server_model = SampleNet(create_model_with_value(0))
        server = instantiate(
            SyncServerConfig(
                aggregation_type=aggregation_type,
                server_optimizer=FedAvgOptimizerConfig(),
            ),
            global_model=server_model,
        )

        client_updates = self._create_client_updates(
            num_clients, aggregation_type=aggregation_type
        )

        for round_num in range(num_rounds):
            server.init_round()
            for delta, weight in zip(client_updates.deltas, client_updates.weights):
                server.receive_update_from_client(
                    Message(
                        model=SampleNet(create_model_with_value(delta)), weight=weight
                    )
                )
            server.step()
            error_msg = model_parameters_equal_to_value(
                server_model, -client_updates.expected_value * (round_num + 1)
            )
            assertEmpty(error_msg, msg=error_msg)

    @pytest.mark.parametrize(
        "aggregation_type",
        [
            AggregationType.AVERAGE,
            AggregationType.WEIGHTED_AVERAGE,
            AggregationType.SUM,
            AggregationType.WEIGHTED_SUM,
        ],
    )
    @pytest.mark.parametrize(
        "num_clients",
        [10, 1],
    )
    @pytest.mark.parametrize(
        "num_rounds",
        [10, 1],
    )
    def test_fed_sgd_sync_server(
        self, aggregation_type, num_clients, num_rounds
    ) -> None:
        opt_config = FedAvgWithLROptimizerConfig(lr=0.1)
        self._compare_optim_and_server(
            opt_config,
            agg_type=aggregation_type,
            num_rounds=num_rounds,
            num_clients=num_clients,
        )

    @pytest.mark.parametrize(
        "aggregation_type",
        [
            AggregationType.AVERAGE,
            AggregationType.WEIGHTED_AVERAGE,
            AggregationType.SUM,
            AggregationType.WEIGHTED_SUM,
        ],
    )
    @pytest.mark.parametrize(
        "num_clients",
        [10, 1],
    )
    @pytest.mark.parametrize(
        "num_rounds",
        [10, 1],
    )
    def test_fed_adam_sync_server(
        self, aggregation_type, num_clients, num_rounds
    ) -> None:
        opt_config = FedAdamOptimizerConfig(lr=0.1)
        self._compare_optim_and_server(
            opt_config,
            agg_type=aggregation_type,
            num_rounds=num_rounds,
            num_clients=num_clients,
        )

    @pytest.mark.parametrize(
        "aggregation_type",
        [
            AggregationType.AVERAGE,
            AggregationType.WEIGHTED_AVERAGE,
            AggregationType.SUM,
            AggregationType.WEIGHTED_SUM,
        ],
    )
    @pytest.mark.parametrize(
        "num_clients",
        [10, 1],
    )
    @pytest.mark.parametrize(
        "num_rounds",
        [10, 1],
    )
    def test_fed_lars_sync_server(
        self, aggregation_type, num_clients, num_rounds
    ) -> None:
        opt_config = FedLARSOptimizerConfig(lr=0.001)
        self._compare_optim_and_server(
            opt_config,
            agg_type=aggregation_type,
            num_rounds=num_rounds,
            num_clients=num_clients,
        )

    @pytest.mark.parametrize(
        "aggregation_type",
        [
            AggregationType.AVERAGE,
            AggregationType.WEIGHTED_AVERAGE,
            AggregationType.SUM,
            AggregationType.WEIGHTED_SUM,
        ],
    )
    @pytest.mark.parametrize(
        "num_clients",
        [10, 1],
    )
    @pytest.mark.parametrize(
        "num_rounds",
        [10, 1],
    )
    def test_fed_lamb_sync_server(
        self, aggregation_type, num_clients, num_rounds
    ) -> None:
        opt_config = FedLAMBOptimizerConfig(lr=0.001)
        self._compare_optim_and_server(
            opt_config,
            agg_type=aggregation_type,
            num_rounds=num_rounds,
            num_clients=num_clients,
        )

    def test_select_clients_for_training(self):
        """
        Selects 10 clients out of 100. SyncServer with seed = 0 should
        return the same indicies as those of uniform random selector.
        """
        selector_config = UniformlyRandomActiveUserSelectorConfig(user_selector_seed=0)
        uniform_selector = instantiate(selector_config)

        server = instantiate(
            SyncServerConfig(active_user_selector=selector_config),
            global_model=SampleNet(create_model_with_value(0)),
        )

        server_selected_indices = server.select_clients_for_training(
            num_total_users=100, users_per_round=10
        )
        uniform_selector_indices = uniform_selector.get_user_indices(
            num_total_users=100, users_per_round=10
        )
        assertEqual(server_selected_indices, uniform_selector_indices)

    @pytest.mark.parametrize(
        "channel",
        [HalfPrecisionChannel(), IdentityChannel()],
    )
    def test_server_channel_integration(self, channel):
        """From Client to Server, the channel should quantize and then dequantize the message
        therefore there should be no change in the model
        """
        server = instantiate(
            SyncServerConfig(),
            global_model=SampleNet(create_model_with_value(0)),
            channel=channel,
        )

        delta = create_model_with_value(1)
        init = copy.deepcopy(delta)
        server.receive_update_from_client(Message(model=SampleNet(delta), weight=1.0))
        error_msg = verify_models_equivalent_after_training(delta, init)
        assertEmpty(error_msg, msg=error_msg)

    def test_sketch_creation_wrong_channel(self):
        """
        Tests if creating SketchRoundReducer with a wrong channel raises an error
        """
        model = SampleNet(TwoFC())
        channel = instantiate(SparseMaskChannelConfig())
        with assertRaises((HydraException, AssertionError)):
            _ = instantiate(
                SketchRoundReducerConfig(),
                global_model=model,
                channel=channel,
            )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, Iterator, Tuple

import torch
from flsim.common.logger import Logger
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from hydra.utils import instantiate
from omegaconf import MISSING
from torch import nn


class FixedPointConverter:
    r"""
    The main class that is responsible for conversion between
    fixed point and floating point.
    """

    MAX_WIDTH = 8
    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(self, **kwargs):
        r"""
        Args:
            cfg: The config for FixedPointConverter

        Raises:
            ValueError: if the ``num_bytes`` is not between 1 and 8, or if
            ``config.scaling_factor`` is not greater than 0.
        """
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=FixedPointConfig,
            **kwargs,
        )

        if self.cfg.num_bytes < 1 or self.cfg.num_bytes > self.MAX_WIDTH:
            error_msg = (
                f"Width {self.cfg.num_bytes} is not supported. "
                f"Please enter a width between 1 and {self.MAX_WIDTH}."
            )
            raise ValueError(error_msg)
        if self.cfg.scaling_factor <= 0:
            raise ValueError("scaling factor must be greater than 0.")
        num_bits = self.cfg.num_bytes * 8
        self.max_value = 2 ** (num_bits - 1) - 1
        self.min_value = -(2 ** (num_bits - 1))
        self.scaling_factor = self.cfg.scaling_factor
        self._overflows = 0
        self._underflows = 0

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def to_fixedpoint(self, numbers: torch.Tensor) -> torch.Tensor:
        """
        Converts numbers in a tensor from floating point to fixed point.

        During conversion, the floats are multiplied by ``scaling_factor``.
        Now if some of these numbers are outside the range that can be represented by
        ``num_bytes`` bytes, they will be clamped to fit in the range.

        Args:
            numbers: the tensor containing the floating point numbers to convert

        Returns:
            A tensor containing the converted numbers to fixed point.
        """

        numbers = numbers.mul(self.scaling_factor)
        if (numbers < self.min_value).any():  # pyre-ignore[16]
            self._underflows = self._underflows + 1
        if (numbers > self.max_value).any():
            self._overflows = self._overflows + 1
        numbers = numbers.clamp(self.min_value, self.max_value)
        return torch.round(numbers)

    def to_float(self, numbers: torch.Tensor) -> torch.Tensor:
        """
        Converts numbers in a tensor from fixed point to floating point.

        Note that this method does not check if the fixed point numbers
        is withing the range of numbers that can be represented by
        ``num_bytes`` bytes.

        Args:
            numbers: the tensor containing the fixed point numbers to convert

        Returns:
            A tensor containing the converted number to floating point.
        """
        return numbers.div(self.scaling_factor)

    def _show_debug_vars(self):
        """
        Shows the underflow and overflow counts if debugging is enabled
        """
        if self.logger.isEnabledFor(logging.DEBUG):
            msg = (
                f"{self._underflows} tensor(s) have underflow and "
                f"{self._overflows} tensor(s) have overflow during fixed point conversion"
            )
            self.logger.warning(msg)


def utility_config_flatter(
    model: nn.Module, flat_config: FixedPointConfig
) -> Dict[str, FixedPointConfig]:
    """
    A utility function to use a "flat" (same config for all layers)
    FixedPointConfig for all layers of a model.

    Args:
        model: the reference model to obtain the named parameters
        flat_config: The flat config to use for all layers

    Returns:
        returns the flat fixedpoint_config_dict
    """
    config: Dict[str, FixedPointConfig] = {}
    for name, _ in model.named_parameters():
        config[name] = flat_config
    return config


class SecureAggregator:
    r"""
    The main class that is responsible for secure aggregation.

    Notes:
        Since this is a simulation of secure aggregation, it is simplified and
        not all details of secure aggregation are implemented. For instance, the
        noise generation, sharing random seed, denoising for secure aggregation
        are not implemented. Also, entities such as secure enclaves are not
        implemented.
    """

    def __init__(
        self,
        config: Dict[str, FixedPointConfig],
    ):
        r"""
        Args:
            config: a dictionary of fixed-point configs for different layers of
                neural network. If the utility ``utility_config_flatter`` is used,
                same config will be used for all layers of the neural network.

        Notes:
            T97188526
        """
        self.converters = {}
        for key in config.keys():
            self.converters[key] = instantiate(config[key])

    def _check_converter_dict_items(self, model: nn.Module) -> None:
        """
        Checks if all layers of a model have their corresponding configs

        Args:
            model: the model

        Raises:
            ValueError: If some layers of the model do not have their
                corresponding configs
        """
        unset_configs = set(model.state_dict()) - set(self.converters)
        if unset_configs:
            error_msg = (
                "Not all "
                "layers have their corresponding fixed point config. "
                f"The layers {unset_configs} do not have configs."
            )
            raise ValueError(error_msg)

    def params_to_fixedpoint(self, model: nn.Module) -> None:
        """
        Converts parameters of a model from floating point to fixed point.

        Args:
            model: the model whose parameters will be converted

        Raises:
            ValueError: If some layers of the model do not have their
                corresponding configs
        """
        self._check_converter_dict_items(model)
        state_dict = model.state_dict()
        for name in state_dict.keys():
            state_dict[name] = self.converters[name].to_fixedpoint(state_dict[name])
            self.converters[name]._show_debug_vars()
        model.load_state_dict(state_dict)

    def params_to_float(self, model: nn.Module) -> None:
        """
        Converts parameters of a model from fixed point to floating point.

        Args:
            model: the model whose parameters will be converted

        Raises:
            ValueError: If some layers of the model do not have their
                corresponding configs
        """
        self._check_converter_dict_items(model)
        state_dict = model.state_dict()
        for name in state_dict.keys():
            state_dict[name] = self.converters[name].to_float(state_dict[name])
        model.load_state_dict(state_dict)

    def _generate_noise_mask(
        self, update_params: Iterator[Tuple[str, nn.Parameter]]
    ) -> Iterator[Tuple[str, nn.Parameter]]:
        """
        Generates noise mask, same shape as the update params

        Args:
            update_params: the parameters of the update sent from
                clients. Used to infer the shape of the noise mask

        Returns:
            noise mask
        """
        pass

    def apply_noise_mask(
        self, update_params: Iterator[Tuple[str, nn.Parameter]]
    ) -> None:
        """
        Applies noise mask to the parameters of the update sent from
            clients.

        Args:
            update_params: the parameters of the update sent from
                clients.

        Note:
            To properly implement this method, call ``_generate_noise_mask()``
            as ``noise_mask = self._generate_noise_mask(update_params)``. Then
            add the ``noise_mask`` to ``update_params`` and return the new
            ``update_params``.
        """
        pass

    def _get_denoise_mask(self) -> Iterator[Tuple[str, nn.Parameter]]:
        """
        Gets the aggregated denoised mask for all participating clients
        from secure enclave.

        Returns:
            aggregated denoised mask for all participating clients
        """
        pass

    def apply_denoise_mask(
        self, model_aggregate_params: Iterator[Tuple[str, nn.Parameter]]
    ) -> None:
        """
        Applies denoise mask to the noised aggregated updates from clients

        Args:
            model_aggregate_params: the parameters of the noised aggragated
                client updates. Used to infer the shape of the denoise mask

        Note:
            To properly implement this method, call ``_get_denoise_mask()``
            as ``denoise_mask = self._get_denoise_mask()``. Then add the
            ``denoise_mask`` to ``model_aggregate_params`` and return the
            new ``model_aggregate_params``.
        """
        pass


@dataclass
class FixedPointConfig:
    _target_: str = fullclassname(FixedPointConverter)
    _recursive_: bool = False
    # size in bytes of single fixed point number. 1 to 8 inclusive.
    num_bytes: int = MISSING
    # multiplier to convert from floating to fixed point
    scaling_factor: int = MISSING



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore

from .base_client import ClientConfig
from .dp_client import DPClientConfig


ConfigStore.instance().store(
    name="base_client",
    node=ClientConfig,
    group="client",
)

ConfigStore.instance().store(
    name="base_dp_client",
    node=DPClientConfig,
    group="client",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

# The base hydra configs for different components of FLSim are defined
# in their respective packages' __init__.py files.
# Importing the trainer module will execute flsim.trainers' __init__
# and cause chain reaction leading to execution and consequent import
# of all the config store definitions. Similarly, importing data module
# will import all necessary dataloader configs.

import flsim.data  # noqa
import flsim.trainers  # noqa



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict

import torch
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from omegaconf import MISSING
from torch.nn import Module as Model  # @manual


class LocalOptimizer:
    def __init__(self, *, model: Model, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=LocalOptimizerConfig,
            **kwargs,
        )
        self.model = model

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass


# pyre-ignore[11] Annotation `torch.optim.SGD` is not defined as a type
class LocalOptimizerSGD(LocalOptimizer, torch.optim.SGD):
    def __init__(self, *, model: Model, **kwargs) -> None:
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=LocalOptimizerSGDConfig,
            **kwargs,
        )

        super().__init__(model=model, **kwargs)

        torch.optim.SGD.__init__(
            self=self,
            params=self.model.parameters(),
            # pyre-fixme[16]: `LocalOptimizerSGD` has no attribute `cfg`.
            lr=self.cfg.lr,
            momentum=self.cfg.momentum,
            weight_decay=self.cfg.weight_decay,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @staticmethod
    def dict_config(
        lr: float = 0.001, momentum: float = 0.0, weight_decay: float = 0.0
    ) -> Dict[str, Any]:
        """Allows downstream functions to get configs given lr and momentum
        With this function, we can change implementation of
        LocalSGDOptimizer.dict_config without changing downstream code
        """
        return {
            "_target_": LocalOptimizerSGDConfig._target_,
            "lr": lr,
            "momentum": momentum,
            "weight_decay": weight_decay,
        }


class LocalOptimizerFedProx(LocalOptimizer, torch.optim.SGD):
    def __init__(
        self,
        *,
        model: Model,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=LocalOptimizerFedProxConfig,
            **kwargs,
        )

        super().__init__(model=model, **kwargs)

        torch.optim.SGD.__init__(
            self=self,
            params=self.model.parameters(),
            # pyre-fixme[16]: `LocalOptimizerFedProx` has no attribute `cfg`.
            lr=self.cfg.lr,
            momentum=self.cfg.momentum,
            weight_decay=self.cfg.weight_decay,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """

        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            weight_decay = group["weight_decay"]
            momentum = group["momentum"]
            dampening = group["dampening"]
            nesterov = group["nesterov"]

            for p in group["params"]:
                if p.grad is None:
                    continue
                d_p = p.grad
                param_state = self.state[p]

                if "global_model" not in param_state:
                    param_state["global_model"] = torch.clone(p.data).detach()

                if weight_decay != 0:
                    d_p = d_p.add(p, alpha=weight_decay)
                if momentum != 0:
                    if "momentum_buffer" not in param_state:
                        buf = param_state["momentum_buffer"] = torch.clone(d_p).detach()
                    else:
                        buf = param_state["momentum_buffer"]
                        buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if nesterov:
                        d_p = d_p.add(buf, alpha=momentum)
                    else:
                        d_p = buf

                d_p.add_(p.data - param_state["global_model"], alpha=self.cfg.mu)
                p.add_(d_p, alpha=-group["lr"])

        return loss


@dataclass
class LocalOptimizerConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    lr: float = 0.001
    momentum: float = 0.0
    weight_decay: float = 0.0


@dataclass
class LocalOptimizerSGDConfig(LocalOptimizerConfig):
    _target_: str = fullclassname(LocalOptimizerSGD)


@dataclass
class LocalOptimizerFedProxConfig(LocalOptimizerConfig):
    _target_: str = fullclassname(LocalOptimizerFedProx)
    mu: float = 0.0



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import flsim.configs  # noqa
import omegaconf
from flsim.common.pytest_helper import assertRaises
from flsim.utils.sample_model import DummyAlphabetFLModel
from hydra.experimental import compose, initialize
from hydra.utils import instantiate


class TestConfigValidation:
    def test_throw_exception_on_missing_field(self) -> None:
        with initialize(config_path=None):
            cfg = compose(
                config_name=None,
                overrides=[
                    "+trainer=base_async_trainer",
                    # ThresholdStalenessWeightConfig has MISSING fields.
                    # Hydra should throw an exception at instantiation time
                    # since we won't be setting those fields here.
                    "+staleness_weight@trainer.async_weight.staleness_weight=base_threshold_staleness_weight",
                ],
            )
        with assertRaises(
            omegaconf.errors.MissingMandatoryValue,  # with Hydra 1.1
        ):
            instantiate(cfg.trainer, model=DummyAlphabetFLModel(), cuda_enabled=False)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore  #  @manual

from .training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
    PerUserGaussianDurationDistributionConfig,
    PerUserHalfNormalDurationDistributionConfig,
    PerUserUniformDurationDistributionConfig,
    PerUserExponentialDurationDistributionConfig,
    DurationDistributionFromListConfig,
)


ConfigStore.instance().store(
    name="base_per_example_gaussian_duration_distribution",
    node=PerExampleGaussianDurationDistributionConfig,
    group="duration_distribution_generator",
)


ConfigStore.instance().store(
    name="base_per_user_gaussian_duration_distribution",
    node=PerUserGaussianDurationDistributionConfig,
    group="duration_distribution_generator",
)


ConfigStore.instance().store(
    name="base_per_user_half_normal_duration_distribution",
    node=PerUserHalfNormalDurationDistributionConfig,
    group="duration_distribution_generator",
)


ConfigStore.instance().store(
    name="base_per_user_uniform_duration_distribution",
    node=PerUserUniformDurationDistributionConfig,
    group="duration_distribution_generator",
)


ConfigStore.instance().store(
    name="base_per_user_exponential_duration_distribution",
    node=PerUserExponentialDurationDistributionConfig,
    group="duration_distribution_generator",
)


ConfigStore.instance().store(
    name="base_duration_distribution_from_list",
    node=DurationDistributionFromListConfig,
    group="duration_distribution_generator",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from enum import Enum, auto

from flsim.utils.fl.stats import RandomVariableStatsTracker


class ChannelDirection(Enum):
    CLIENT_TO_SERVER = auto()
    SERVER_TO_CLIENT = auto()


class ChannelStatsCollector:
    def __init__(self):
        self.reset_channel_stats()

    def reset_channel_stats(self):
        self.communication_stats = {
            ChannelDirection.CLIENT_TO_SERVER: RandomVariableStatsTracker(),
            ChannelDirection.SERVER_TO_CLIENT: RandomVariableStatsTracker(),
        }

    def get_channel_stats(self):
        return self.communication_stats

    def collect_channel_stats(
        self, message_size_bytes: float, client_to_server: bool = True
    ):
        """
        Collect statistics about the updates/model transmitted both
        for client to server and server to client directions.
        """
        direction = (
            ChannelDirection.CLIENT_TO_SERVER
            if client_to_server
            else ChannelDirection.SERVER_TO_CLIENT
        )
        self.communication_stats[direction].update(message_size_bytes)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""
This file contains the functions for clipping clients' updates in
an FL simulation.
"""

from typing import Optional

import torch
from flsim.utils.fl.common import FLModelParamUtils
from torch import nn


__EPS__ = 1e-10


class UserUpdateClipper:
    def __init__(self, precision: Optional[torch.dtype] = None):
        self._cached_model_diff = None
        self.precision = precision

    def calc_model_diff(self, new_model: nn.Module, prev_model: nn.Module) -> nn.Module:
        """
        Calculates the difference between the updated model and the old model
        """
        if self._cached_model_diff is None:  # for memory efficiency purposes
            self._cached_model_diff = FLModelParamUtils.clone(new_model, self.precision)
        FLModelParamUtils.linear_comb_models(
            new_model, 1, prev_model, -1, self._cached_model_diff
        )
        return self._cached_model_diff

    def clip(self, model_diff: nn.Module, max_norm: float) -> None:
        """
        Clips user update (stored in ``model_diff``) by computing clip factor
        and using it to rescale each user's update (operation is in-place).

        This method clips the parameters of the user update. This operation
        is in-place (modifies ``model_diff`` in this method)
        """
        max_norm = float(max_norm)  # TODO: (extension to per-layer clip) T69971799

        per_user_update_norm = self._calc_norm(model_diff.parameters())
        clip_factor = self._calc_clip_factor(max_norm, per_user_update_norm)

        with torch.no_grad():
            for parameter in model_diff.parameters():
                parameter.copy_(parameter * clip_factor)

    def _calc_clip_factor(self, max_norm: float, per_user_norm: float):
        """
        Calculates the clip factor that will be used to clip the user updatas
        """
        if max_norm < 0 or per_user_norm < 0:
            raise ValueError("Error: max_norm and per_user_norm must be both positive.")
        clip_factor = max_norm / (per_user_norm + __EPS__)
        clip_factor = min(clip_factor, 1.0)
        return clip_factor  # TODO: (consolidate with another implementation) T69972084

    def _calc_norm(self, params):
        """
        Calculates the l-2 norm of the user updates
        """
        norms = [param.norm(2) for param in params]
        norm = torch.tensor(norms).norm(2)
        return norm



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Iterable

import torch
import torch.nn as nn
from flsim.utils.fl.personalized_model import FLModelWithPrivateModules


class LinearRegression(nn.Module):
    def __init__(self):
        super().__init__()
        self.a = nn.Parameter(torch.randn(1, requires_grad=True, dtype=torch.float))
        self.b = nn.Parameter(torch.randn(1, requires_grad=True, dtype=torch.float))

    def forward(self, x):
        return self.a + self.b * x


LR_PRIVATE_MODULE_NAMES = {"a"}


class PersonalizedLinearRegression(LinearRegression, FLModelWithPrivateModules):
    @classmethod
    def _get_user_private_module_names(cls) -> Iterable[str]:
        return LR_PRIVATE_MODULE_NAMES

    def __init__(self):
        super().__init__()

        # Set up user-private module attributes whenever we create a new
        # instance.
        self._maybe_set_up_user_private_modules()

        # Set forward hooks to reuse the forward() of the parent class.
        self._set_forward_hooks()


class FCModel(nn.Module):
    def __init__(self):
        super(FCModel, self).__init__()
        self.fc1 = nn.Linear(10, 5)
        self.fc2 = nn.Linear(5, 3)
        self.fc3 = nn.Linear(3, 1)

    def forward(self, x):
        x = self.fc1(x)
        x = self.fc2(x)
        x = self.fc3(x)
        return x


FC_PRIVATE_MODULE_NAMES = {"fc1", "fc3"}


class PersonalizedFCModel(FCModel, FLModelWithPrivateModules):
    @classmethod
    def _get_user_private_module_names(cls) -> Iterable[str]:
        return FC_PRIVATE_MODULE_NAMES

    def __init__(self):
        super().__init__()

        # Set up user-private module attributes whenever we create a new
        # instance.
        self._maybe_set_up_user_private_modules()

        # Set forward hooks to reuse the forward() of the parent class.
        self._set_forward_hooks()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import copy
from dataclasses import dataclass
from typing import List, Tuple

import torch
from flsim.active_user_selectors.simple_user_selector import (
    ActiveUserSelector,
    ActiveUserSelectorConfig,
)
from flsim.common.diversity_metrics import (
    DiversityMetrics,
    DiversityMetricType,
    DiversityStatistics,
)
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.model import IFLModel
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg


class DiverseUserSelectorUtils:
    @staticmethod
    def calculate_diversity_metrics(
        data_provider: IFLDataProvider,
        global_model: IFLModel,
        user_indices: List[int],
        loss_reduction_type: str = "mean",
        client_gradient_scaling: str = "sum",
        diversity_metric_type: DiversityMetricType = DiversityMetricType.gradient_diversity,
    ) -> DiversityMetrics:
        """Calculates the Gradient Diversity of a cohort -- see arXiv:1706.05699
        For the FL application, define a aggregate gradient for each user, which is the sum of
        the gradient of the loss function for all the user's data samples (denoted grad f_i).
        We compute two quantities: norm_of_sum is the norm of the sum of these user
        gradients, and sum_of_norms is the sum of the norms of these gradients. These quantities
        are passed to a DiversityMetrics object which computes the gradient diversity metrics.

        norm_of_sum = || sum grad f_i ||_2^2
        sum_of_norms = sum || grad f_i ||_2^2

        loss_reduction_type should match the loss reduction type of the model (pytorch defaults to mean)
        client_gradient_scaling determines how the gradient for each client is computed. Either
        the gradient is summed over all the user's data, or it is the average client gradient.
        """

        if loss_reduction_type not in {"sum", "mean"}:
            raise ValueError('loss_reduction_type must be "sum" or "mean"')
        if client_gradient_scaling not in {"sum", "mean"}:
            raise ValueError('client_gradient_scaling must be "sum" or "mean"')

        global_model.fl_get_module().train()
        params = list(global_model.fl_get_module().parameters())

        def zero_like_grad(param_list: List[torch.Tensor]) -> List[torch.Tensor]:
            zero_gradient = []
            for group in param_list:
                if group.requires_grad:
                    zero_gradient.append(torch.zeros_like(group, requires_grad=False))
            return zero_gradient

        gradient_sum = zero_like_grad(params)
        norm_of_sum, sum_of_norms = 0.0, 0.0

        # Accumulate the gradient over a user's batches, normalized by batch size
        for user_idx in user_indices:
            user_data = data_provider.get_user_data(user_idx)
            accumulated_gradient = zero_like_grad(params)

            for batch in user_data:
                global_model.fl_get_module().zero_grad()
                batch_metrics = global_model.fl_forward(batch)
                batch_metrics.loss.backward()

                for group_in, group_out in zip(params, accumulated_gradient):
                    if loss_reduction_type == "mean":
                        group_out += group_in.grad * batch_metrics.num_examples
                    else:
                        group_out += group_in.grad

            for group_in, group_out in zip(accumulated_gradient, gradient_sum):
                if client_gradient_scaling == "mean":
                    group_in = group_in / user_data.num_examples()

                group_out += group_in
                sum_of_norms += torch.sum(group_in * group_in).item()

        for group in gradient_sum:
            norm_of_sum += torch.sum(group * group).item()

        return DiversityMetrics(
            norm_of_sum=norm_of_sum,
            sum_of_norms=sum_of_norms,
            diversity_metric_type=diversity_metric_type,
        )

    @staticmethod
    def select_diverse_cohort(
        data_provider: IFLDataProvider,
        global_model: IFLModel,
        users_per_round: int,
        available_users: List[int],
        rng: torch.Generator,
        num_search_samples: int,
        maximize_metric: bool,
        loss_reduction_type: str = "mean",
        client_gradient_scaling: str = "sum",
        diversity_metric_type: DiversityMetricType = DiversityMetricType.gradient_diversity,
    ) -> Tuple[List[int], DiversityStatistics]:
        """Choose a cohort which optimizes the diversity metric type out of some number
        of sample cohorts. Calculate the max, min, and average metric of the sample cohorts.

        Inputs:
        maximize_metric: whether the diversity metric is maximized (True) or minimized (False)

        Outputs:
        Cohort (out of the sample cohorts) which maximizes the metric
        Statistics of the metric of interest of the cohort.
        """

        if users_per_round >= len(available_users):
            candidate_selected_indices = copy.copy(available_users)
            candidate_metrics = DiverseUserSelectorUtils.calculate_diversity_metrics(
                data_provider=data_provider,
                global_model=global_model,
                user_indices=candidate_selected_indices,
                loss_reduction_type=loss_reduction_type,
                client_gradient_scaling=client_gradient_scaling,
                diversity_metric_type=diversity_metric_type,
            )
            cohort_stats = DiversityStatistics([candidate_metrics])

            return candidate_selected_indices, cohort_stats

        # Choose cohorts from the available users and calculate their diversity statistics.
        selected_indices = torch.multinomial(
            torch.ones(len(available_users), dtype=torch.float),
            users_per_round,
            replacement=False,
            generator=rng,
        ).tolist()

        candidate_user_indices = [available_users[idx] for idx in selected_indices]

        candidate_diversity_metrics = (
            DiverseUserSelectorUtils.calculate_diversity_metrics(
                data_provider=data_provider,
                global_model=global_model,
                user_indices=candidate_user_indices,
                loss_reduction_type=loss_reduction_type,
                client_gradient_scaling=client_gradient_scaling,
                diversity_metric_type=diversity_metric_type,
            )
        )

        sample_cohort_metrics = [candidate_diversity_metrics]

        for _ in range(num_search_samples - 1):
            selected_indices = torch.multinomial(
                torch.ones(len(available_users), dtype=torch.float),
                users_per_round,
                replacement=False,
                generator=rng,
            ).tolist()

            user_indices = [available_users[idx] for idx in selected_indices]
            cohort_diversity_metrics = (
                DiverseUserSelectorUtils.calculate_diversity_metrics(
                    data_provider=data_provider,
                    global_model=global_model,
                    user_indices=user_indices,
                    loss_reduction_type=loss_reduction_type,
                    client_gradient_scaling=client_gradient_scaling,
                    diversity_metric_type=diversity_metric_type,
                )
            )

            sample_cohort_metrics.append(cohort_diversity_metrics)

            if maximize_metric and (
                cohort_diversity_metrics > candidate_diversity_metrics
            ):
                candidate_user_indices = user_indices
                candidate_diversity_metrics = cohort_diversity_metrics
            elif (not maximize_metric) and (
                cohort_diversity_metrics < candidate_diversity_metrics
            ):
                candidate_user_indices = user_indices
                candidate_diversity_metrics = cohort_diversity_metrics

        cohort_stats = DiversityStatistics(sample_cohort_metrics)

        return candidate_user_indices, cohort_stats


class DiversityReportingUserSelector(ActiveUserSelector):
    """User Selector which chooses users uniformly randomly and records
    the Gradient Diversity (GD) of the cohort. For GD, see arXiv:1706.05699.

    If constant_cohorts is set to True, choose a set of cohorts at the start of
    training and at each round, report the GD of each of those sample cohorts.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=DiversityReportingUserSelectorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

        if not isinstance(self.cfg.diversity_metric_type, DiversityMetricType):
            raise ValueError("diversity_metric_type must be of DiversityMetricType")
        if self.cfg.loss_reduction_type not in {"sum", "mean"}:
            raise ValueError('loss_reduction_type must be "sum" or "mean"')
        if self.cfg.client_gradient_scaling not in {"sum", "mean"}:
            raise ValueError('client_gradient_scaling must be "sum" or "mean"')

        self.sample_cohorts = []

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def get_user_indices(self, **kwargs) -> List[int]:
        required_inputs = [
            "num_total_users",
            "users_per_round",
            "data_provider",
            "global_model",
            "epoch",
        ]
        (
            num_total_users,
            users_per_round,
            data_provider,
            global_model,
            epoch,
        ) = self.unpack_required_inputs(required_inputs, kwargs)

        selected_indices = torch.multinomial(
            torch.ones(num_total_users, dtype=torch.float),
            users_per_round,
            replacement=False,
            generator=self.rng,
        ).tolist()

        currently_active = (
            # pyre-fixme[16]: `DiversityReportingUserSelector` has no attribute `cfg`.
            epoch >= self.cfg.epochs_before_active
            and epoch < self.cfg.epochs_before_active + self.cfg.num_epochs_active
        )

        if not currently_active:
            return selected_indices

        if len(self.sample_cohorts) == 0 or not self.cfg.constant_cohorts:
            self.sample_cohorts = []
            for _ in range(self.cfg.num_candidate_cohorts):
                sample_cohort_indices = torch.multinomial(
                    torch.ones(num_total_users, dtype=torch.float),
                    users_per_round,
                    replacement=False,
                    generator=self.rng,
                ).tolist()

                self.sample_cohorts.append(sample_cohort_indices)

        diversity_metrics_list = []
        for sample_cohort in self.sample_cohorts:
            diversity_metrics_list.append(
                DiverseUserSelectorUtils.calculate_diversity_metrics(
                    data_provider=data_provider,
                    global_model=global_model,
                    user_indices=sample_cohort,
                    loss_reduction_type=self.cfg.loss_reduction_type,
                    client_gradient_scaling=self.cfg.client_gradient_scaling,
                    diversity_metric_type=self.cfg.diversity_metric_type,
                )
            )

        return selected_indices


class DiversityStatisticsReportingUserSelector(ActiveUserSelector):
    """Find the statistics of the diversity metric of interest.
    The cohort is selected uniformly randomly with replacement.
    Record the sample maximum, average, and minimum metric for:
        1) cohorts chosen with replacement
        2) cohorts chosen without replacement (round-robin), where the cohort
        with the maximum metric is removed from the list of available users.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=DiversityStatisticsReportingUserSelectorConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

        if not isinstance(self.cfg.diversity_metric_type, DiversityMetricType):
            raise ValueError("diversity_metric_type must be of DiversityMetricType")
        if self.cfg.loss_reduction_type not in {"sum", "mean"}:
            raise ValueError('loss_reduction_type must be "sum" or "mean"')
        if self.cfg.client_gradient_scaling not in {"sum", "mean"}:
            raise ValueError('client_gradient_scaling must be "sum" or "mean"')

        self.available_users = []
        self.cohort_stats_with_replacement = []
        self.cohort_stats_without_replacement = []

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def get_user_indices(self, **kwargs) -> List[int]:
        required_inputs = [
            "num_total_users",
            "users_per_round",
            "data_provider",
            "global_model",
            "epoch",
        ]
        (
            num_total_users,
            users_per_round,
            data_provider,
            global_model,
            epoch,
        ) = self.unpack_required_inputs(required_inputs, kwargs)

        if len(self.available_users) == 0:
            self.available_users = list(range(num_total_users))

        # First uniformly randomly select the actual cohort used for training
        baseline_selected_indices = torch.multinomial(
            torch.ones(num_total_users, dtype=torch.float),
            users_per_round,
            replacement=False,
            generator=self.rng,
        ).tolist()

        if (
            # pyre-fixme[16]: `DiversityStatisticsReportingUserSelector` has no
            #  attribute `cfg`.
            epoch < self.cfg.epochs_before_active
            or epoch >= self.cfg.epochs_before_active + self.cfg.num_epochs_active
        ):
            return baseline_selected_indices

        (_, stat_with_replacement) = DiverseUserSelectorUtils.select_diverse_cohort(
            data_provider=data_provider,
            global_model=global_model,
            users_per_round=users_per_round,
            available_users=list(range(num_total_users)),
            rng=self.rng,
            num_search_samples=self.cfg.num_candidate_cohorts,
            maximize_metric=self.cfg.maximize_metric,
            loss_reduction_type=self.cfg.loss_reduction_type,
            client_gradient_scaling=self.cfg.client_gradient_scaling,
            diversity_metric_type=self.cfg.diversity_metric_type,
        )

        (
            candidate_user_indices,
            stat_without_replacement,
        ) = DiverseUserSelectorUtils.select_diverse_cohort(
            data_provider=data_provider,
            global_model=global_model,
            users_per_round=users_per_round,
            available_users=self.available_users,
            rng=self.rng,
            num_search_samples=self.cfg.num_candidate_cohorts,
            maximize_metric=self.cfg.maximize_metric,
            loss_reduction_type=self.cfg.loss_reduction_type,
            client_gradient_scaling=self.cfg.client_gradient_scaling,
            diversity_metric_type=self.cfg.diversity_metric_type,
        )

        self.cohort_stats_with_replacement.append(stat_with_replacement)
        self.cohort_stats_without_replacement.append(stat_without_replacement)

        # Update the list of available users for round-robin selector
        self.available_users = [
            idx for idx in self.available_users if idx not in candidate_user_indices
        ]

        return baseline_selected_indices


class DiversityMaximizingUserSelector(ActiveUserSelector):
    """A user selector class which chooses cohorts to greedily maximize their Diversity.

    Config options:
    num_candidate_cohorts: The number of cohorts considered; the cohort with maximum Diversity
    loss_reduction_type: This should match the loss reduction type in the learning model used
    client_gradient_scaling: How to compute the gradient summary of a client -- whether is should
        be the sum or average of the gradients of its data.
    diversity_metric_type: What is the Diversity Metric which is being optimized
    maximize_metric: Whether to maximize (True) or minimize (False) the specified diversity metric
    with_replacement: If false, choose cohorts in a round-robin fashion. If true, select each cohort
        from all clients at each round.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=DiversityMaximizingUserSelectorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

        if not isinstance(self.cfg.diversity_metric_type, DiversityMetricType):
            raise ValueError("diversity_metric_type must be of DiversityMetricType")
        if self.cfg.loss_reduction_type not in {"sum", "mean"}:
            raise ValueError('loss_reduction_type must be "sum" or "mean"')
        if self.cfg.client_gradient_scaling not in {"sum", "mean"}:
            raise ValueError('client_gradient_scaling must be "sum" or "mean"')

        self.available_users = []

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def get_user_indices(self, **kwargs) -> List[int]:
        required_inputs = [
            "num_total_users",
            "users_per_round",
            "data_provider",
            "global_model",
            "epoch",
        ]
        (
            num_total_users,
            users_per_round,
            data_provider,
            global_model,
            epoch,
        ) = self.unpack_required_inputs(required_inputs, kwargs)

        if len(self.available_users) == 0:
            self.available_users = list(range(num_total_users))

        if (
            # pyre-fixme[16]: `DiversityMaximizingUserSelector` has no attribute `cfg`.
            epoch < self.cfg.epochs_before_active
            or epoch >= self.cfg.epochs_before_active + self.cfg.num_epochs_active
        ):
            baseline_selected_indices = torch.multinomial(
                torch.ones(num_total_users, dtype=torch.float),
                users_per_round,
                replacement=False,
                generator=self.rng,
            ).tolist()
            return baseline_selected_indices

        (candidate_user_indices, _,) = DiverseUserSelectorUtils.select_diverse_cohort(
            data_provider=data_provider,
            global_model=global_model,
            users_per_round=users_per_round,
            available_users=self.available_users,
            rng=self.rng,
            num_search_samples=self.cfg.num_candidate_cohorts,
            maximize_metric=self.cfg.maximize_metric,
            loss_reduction_type=self.cfg.loss_reduction_type,
            client_gradient_scaling=self.cfg.client_gradient_scaling,
            diversity_metric_type=self.cfg.diversity_metric_type,
        )

        if not self.cfg.with_replacement:
            self.available_users = [
                idx for idx in self.available_users if idx not in candidate_user_indices
            ]

        return candidate_user_indices


@dataclass
class DiversityReportingUserSelectorConfig(ActiveUserSelectorConfig):
    _target_: str = fullclassname(DiversityReportingUserSelector)
    diversity_metric_type: DiversityMetricType = DiversityMetricType.gradient_diversity
    epochs_before_active: int = 0
    num_epochs_active: int = int(10e8)
    num_candidate_cohorts: int = 10
    loss_reduction_type: str = "mean"
    client_gradient_scaling: str = "sum"
    constant_cohorts: bool = False


@dataclass
class DiversityStatisticsReportingUserSelectorConfig(ActiveUserSelectorConfig):
    _target_: str = fullclassname(DiversityStatisticsReportingUserSelector)
    diversity_metric_type: DiversityMetricType = DiversityMetricType.gradient_diversity
    maximize_metric: bool = True
    epochs_before_active: int = 0
    num_epochs_active: int = int(10e8)
    num_candidate_cohorts: int = 10
    loss_reduction_type: str = "mean"
    client_gradient_scaling: str = "sum"


@dataclass
class DiversityMaximizingUserSelectorConfig(ActiveUserSelectorConfig):
    _target_: str = fullclassname(DiversityMaximizingUserSelector)
    diversity_metric_type: DiversityMetricType = DiversityMetricType.gradient_diversity
    maximize_metric: bool = True
    epochs_before_active: int = 0
    num_epochs_active: int = int(10e8)
    num_candidate_cohorts: int = 10
    loss_reduction_type: str = "mean"
    client_gradient_scaling: str = "sum"
    with_replacement: bool = False



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import numpy as np
from flsim.utils.async_trainer.async_example_weights import (
    ExampleWeight,
    AsyncExampleWeightConfig,
)
from flsim.utils.async_trainer.async_staleness_weights import (
    AsyncStalenessWeightConfig,
    StalenessWeight,
    ConstantStalenessWeightConfig,
    ThresholdStalenessWeightConfig,
    PolynomialStalenessWeightConfig,
)
from flsim.utils.async_trainer.async_weights import AsyncWeightConfig
from flsim.utils.tests.helpers.async_weights_test_utils import (
    AsyncExampleWeightsTestUtils,
    AsyncStalenessWeightsTestUtils,
)
from hydra.utils import instantiate
from libfb.py import testutil


class AsyncExampleWeightsTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    # two data_providers together produce a cartesian product
    @testutil.data_provider(AsyncExampleWeightsTestUtils.provide_example_weight_configs)
    @testutil.data_provider(
        AsyncStalenessWeightsTestUtils.provide_staleness_weight_configs
    )
    def test_string_conversion(
        self,
        example_weight_config: AsyncExampleWeightConfig,
        example_weight_class: ExampleWeight,
        staleness_weight_config: AsyncStalenessWeightConfig,
        staleness_weight_class: StalenessWeight,
    ) -> None:
        """Check that strings are correctly converted to AsyncWeight"""
        obj = instantiate(
            AsyncWeightConfig(
                staleness_weight=staleness_weight_config,
                example_weight=example_weight_config,
            )
        )
        self.assertEqual(obj.example_weight.__class__, example_weight_class)
        self.assertEqual(obj.staleness_weight.__class__, staleness_weight_class)

    @testutil.data_provider(AsyncExampleWeightsTestUtils.provide_example_weight_configs)
    def test_weight_compute(
        self,
        example_weight_config: AsyncExampleWeightConfig,
        example_weight_class: ExampleWeight,
        avg_num_examples: int = 1,
        avg_staleness: int = 1,
    ):
        """Test that all weight computation works as expected"""
        max_num_examples = 10000
        max_staleness = 10000
        cutoff = 5000
        value_after_cutoff = 0.001
        exponent = 0.5
        # dict below tells us how to initialize weight object for different
        # staleness weight types
        staleness_weight_configs = [
            ConstantStalenessWeightConfig(),
            ThresholdStalenessWeightConfig(
                cutoff=cutoff, value_after_cutoff=value_after_cutoff
            ),
            PolynomialStalenessWeightConfig(exponent=exponent),
        ]
        for staleness_weight_config in staleness_weight_configs:
            staleness_weight_obj = instantiate(staleness_weight_config)
            # for 10 random integers
            for _ in range(10):
                num_examples = np.random.randint(1, max_num_examples)

                staleness = np.random.randint(1, max_staleness)
                staleness_weight = staleness_weight_obj.weight(staleness)

                example_weight_config.avg_num_examples = avg_num_examples
                example_weight_obj = instantiate(example_weight_config)
                example_weight = example_weight_obj.weight(num_examples)

                expected_combined_weight = example_weight * staleness_weight
                combined_weight_object = instantiate(
                    AsyncWeightConfig(
                        example_weight=example_weight_config,
                        staleness_weight=staleness_weight_config,
                    )
                )
                combined_weight = combined_weight_object.weight(
                    num_examples=num_examples, staleness=staleness
                )
                self.assertEqual(expected_combined_weight, combined_weight)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""In this tutorial, we will train a binary sentiment classifier on LEAF's Sent140 dataset with FLSim.

Before running this file, you need to download the dataset, and partition the data by users. We
provided the script get_data.sh for such task.

  Typical usage example:

  python3 sent140_tutorial.py -- \\
  --config-file fblearner/flow/projects/papaya/examples/hydra_configs/sent140_single_process.json
"""
import itertools
import json
import re
import string
import unicodedata
from typing import List

import flsim.configs  # noqa
import hydra  # @manual
import torch
import torch.nn as nn
from flsim.examples.data.data_providers import LEAFDataLoader, LEAFDataProvider
from flsim.examples.metrics_reporter.fl_metrics_reporter import MetricsReporter
from flsim.examples.models.cv_model import FLModel
from flsim.interfaces.metrics_reporter import Channel
from flsim.utils.config_utils import maybe_parse_json_config
from hydra.utils import instantiate
from omegaconf import DictConfig, OmegaConf
from torch.utils.data import Dataset


class CharLSTM(nn.Module):
    def __init__(
        self,
        num_classes,
        n_hidden,
        num_embeddings,
        embedding_dim,
        max_seq_len,
        dropout_rate,
    ):
        super().__init__()
        self.dropout_rate = dropout_rate
        self.n_hidden = n_hidden
        self.num_classes = num_classes
        self.max_seq_len = max_seq_len
        self.num_embeddings = num_embeddings

        self.embedding = nn.Embedding(
            num_embeddings=self.num_embeddings, embedding_dim=embedding_dim
        )
        self.lstm = nn.LSTM(
            input_size=embedding_dim,
            hidden_size=self.n_hidden,
            num_layers=2,
            batch_first=True,
            dropout=self.dropout_rate,
        )
        self.fc = nn.Linear(self.n_hidden, self.num_classes)
        self.dropout = nn.Dropout(p=self.dropout_rate)

    def forward(self, x):
        seq_lens = torch.sum(x != (self.num_embeddings - 1), 1) - 1
        x = self.embedding(x)  # [B, S] -> [B, S, E]
        out, _ = self.lstm(x)  # [B, S, E] -> [B, S, H]
        out = out[torch.arange(out.size(0)), seq_lens]
        out = self.fc(self.dropout(out))  # [B, S, H] -> # [B, S, C]
        return out


class Sent140Dataset(Dataset):
    def __init__(self, data_root, max_seq_len):
        self.data_root = data_root
        self.max_seq_len = max_seq_len
        self.all_letters = {c: i for i, c in enumerate(string.printable)}
        self.num_letters = len(self.all_letters)
        self.UNK: int = self.num_letters

        with open(data_root, "r+") as f:
            self.dataset = json.load(f)

        self.data = {}
        self.targets = {}

        self.num_classes = 2

        # Populate self.data and self.targets
        for user_id, user_data in self.dataset["user_data"].items():
            self.data[user_id] = self.process_x(list(user_data["x"]))
            self.targets[user_id] = self.process_y(list(user_data["y"]))

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        for user_id in self.data.keys():
            yield self.__getitem__(user_id)

    def __getitem__(self, user_id: str):
        if user_id not in self.data or user_id not in self.targets:
            raise IndexError(f"User {user_id} is not in dataset")

        return self.data[user_id], self.targets[user_id]

    def unicodeToAscii(self, s):
        return "".join(
            c
            for c in unicodedata.normalize("NFD", s)
            if unicodedata.category(c) != "Mn" and c in self.all_letters
        )

    def line_to_indices(self, line: str, max_seq_len: int):
        line_list = self.split_line(line)  # split phrase in words
        line_list = line_list
        chars = self.flatten_list([list(word) for word in line_list])
        # padding
        indices: List[int] = [
            self.all_letters.get(letter, self.UNK)
            for i, letter in enumerate(chars)
            if i < max_seq_len
        ]
        indices = indices + ([self.UNK] * (max_seq_len - len(indices)))
        return indices

    def process_x(self, raw_x_batch):
        x_batch = [e[4] for e in raw_x_batch]
        x_batch = [self.line_to_indices(e, self.max_seq_len) for e in x_batch]
        x_batch = torch.LongTensor(x_batch)
        return x_batch

    def process_y(self, raw_y_batch):
        y_batch = [int(e) for e in raw_y_batch]
        return y_batch

    def split_line(self, line):
        """split given line/phrase into list of words
        Args:
            line: string representing phrase to be split

        Return:
            list of strings, with each string representing a word
        """
        return re.findall(r"[\w']+|[.,!?;]", line)

    def flatten_list(self, nested_list):
        return list(itertools.chain.from_iterable(nested_list))


def build_data_provider(data_config, drop_last=False):

    train_dataset = Sent140Dataset(
        data_root="train/all_data_0_01_keep_1_train_9.json",
        max_seq_len=data_config.max_seq_len,
    )
    test_dataset = Sent140Dataset(
        data_root="test/all_data_0_01_keep_1_test_9.json",
        max_seq_len=data_config.max_seq_len,
    )

    dataloader = LEAFDataLoader(
        train_dataset,
        test_dataset,
        test_dataset,
        batch_size=data_config.local_batch_size,
        drop_last=drop_last,
    )

    data_provider = LEAFDataProvider(dataloader)
    return data_provider, train_dataset.num_letters


def main_worker(
    trainer_config,
    model_config,
    data_config,
    use_cuda_if_available=True,
    rank=0,
    distributed_world_size=1,
):
    data_provider, num_letters = build_data_provider(data_config)

    model = CharLSTM(
        num_classes=model_config.num_classes,
        n_hidden=model_config.n_hidden,
        num_embeddings=num_letters + 1,
        embedding_dim=100,
        max_seq_len=data_config.max_seq_len,
        dropout_rate=model_config.dropout_rate,
    )
    cuda_enabled = torch.cuda.is_available() and use_cuda_if_available
    device = torch.device(f"cuda:{0}" if cuda_enabled else "cpu")
    global_model = FLModel(model, device)
    if cuda_enabled:
        global_model.fl_cuda()
    trainer = instantiate(trainer_config, model=global_model, cuda_enabled=cuda_enabled)

    metrics_reporter = MetricsReporter([Channel.TENSORBOARD, Channel.STDOUT])

    final_model, eval_score = trainer.train(
        data_provider=data_provider,
        metric_reporter=metrics_reporter,
        num_total_users=data_provider.num_users(),
        distributed_world_size=distributed_world_size,
        rank=rank,
    )

    trainer.test(
        data_iter=data_provider.test_data(),
        metric_reporter=MetricsReporter([Channel.STDOUT]),
    )


@hydra.main(config_path=None, config_name="sent140_single_process")
def run(cfg: DictConfig):
    print(OmegaConf.to_yaml(cfg))
    assert cfg.distributed_world_size == 1, "Distributed training is not yet supported."

    trainer_config = cfg.trainer
    model_config = cfg.model
    data_config = cfg.data

    main_worker(trainer_config, model_config, data_config)


if __name__ == "__main__":
    cfg = maybe_parse_json_config()
    run(cfg)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore  #  @manual

from .async_example_weights import (
    Log10ExampleWeightConfig,
    SqrtExampleWeightConfig,
    LinearExampleWeightConfig,
    EqualExampleWeightConfig,
)
from .async_staleness_weights import (
    ConstantStalenessWeightConfig,
    ThresholdStalenessWeightConfig,
    PolynomialStalenessWeightConfig,
)
from .async_weights import AsyncWeightConfig
from .training_event_generator import (
    PoissonAsyncTrainingStartTimeDistrConfig,
    ConstantAsyncTrainingStartTimeDistrConfig,
    AsyncTrainingEventGeneratorFromListConfig,
    AsyncTrainingEventGeneratorConfig,
)


ConfigStore.instance().store(
    name="base_log10_example_weight",
    node=Log10ExampleWeightConfig,
    group="example_weight",
)


ConfigStore.instance().store(
    name="base_sqrt_example_weight",
    node=SqrtExampleWeightConfig,
    group="example_weight",
)


ConfigStore.instance().store(
    name="base_linear_example_weight",
    node=LinearExampleWeightConfig,
    group="example_weight",
)


ConfigStore.instance().store(
    name="base_equal_example_weight",
    node=EqualExampleWeightConfig,
    group="example_weight",
)


ConfigStore.instance().store(
    name="base_constant_staleness_weight",
    node=ConstantStalenessWeightConfig,
    group="staleness_weight",
)


ConfigStore.instance().store(
    name="base_threshold_staleness_weight",
    node=ThresholdStalenessWeightConfig,
    group="staleness_weight",
)


ConfigStore.instance().store(
    name="base_polynomial_staleness_weight",
    node=PolynomialStalenessWeightConfig,
    group="staleness_weight",
)


ConfigStore.instance().store(
    name="base_async_weight",
    node=AsyncWeightConfig,
    group="async_weight",
)


ConfigStore.instance().store(
    name="base_poisson_training_start_time_distribution",
    node=PoissonAsyncTrainingStartTimeDistrConfig,
    group="training_start_time_distribution",
)


ConfigStore.instance().store(
    name="base_constant_training_start_time_distribution",
    node=ConstantAsyncTrainingStartTimeDistrConfig,
    group="training_start_time_distribution",
)


ConfigStore.instance().store(
    name="base_async_training_event_generator_from_list",
    node=AsyncTrainingEventGeneratorFromListConfig,
    group="training_event_generator",
)


ConfigStore.instance().store(
    name="base_async_training_event_generator",
    node=AsyncTrainingEventGeneratorConfig,
    group="training_event_generator",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import unittest

import torch
from flsim.utils.data.data_utils import batchify, merge_dicts
from flsim.utils.data.fake_data_utils import FakeDataProvider, FakeUserData


class DataUtilsTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_batchify(self) -> None:
        self.assertEqual(list(batchify([1, 2, 3, 4, 5], 2)), [[1, 2], [3, 4], [5]])
        self.assertEqual(list(batchify([1, 2, 3, 4, 5], 3)), [[1, 2, 3], [4, 5]])
        self.assertEqual(list(batchify([1, 2, 3, 4], 2)), [[1, 2], [3, 4]])
        self.assertEqual(list(batchify([1, 2, 3, 4], 1)), [[1], [2], [3], [4]])

    def test_merge_dicts(self) -> None:
        expected = {"a": torch.Tensor([[1.0], [2.0]])}
        for key, actual in merge_dicts(
            [{"a": torch.Tensor([1])}, {"a": torch.Tensor([2])}]
        ).items():
            self.assertTrue(key in expected)
            self.assertTrue(torch.all(actual.eq(expected[key])))

        expected = {"a": torch.Tensor([[1.0]]), "b": torch.Tensor([[2.0]])}
        for key, actual in merge_dicts(
            [{"a": torch.Tensor([1])}, {"b": torch.Tensor([2])}]
        ).items():
            self.assertTrue(key in expected)
            self.assertTrue(torch.all(actual.eq(expected[key])))

    def user_data_test_util(
        self,
        user_dataset,
        expected_num_examples,
        expected_batch_size,
        expected_num_batches,
    ):
        self.assertEqual(user_dataset.num_examples(), expected_num_examples)

        for i, batch in enumerate(user_dataset):
            self.assertLessEqual(len(batch["data"]), expected_batch_size)
            last_batch = i
        self.assertEqual(last_batch + 1, expected_num_batches)

    def test_fake_user_data(self):
        def gen_batch(n, value=None):
            return {"data": [torch.ones(n, 10)], "label": [1] * n}

        num_examples = 100
        batch_size = 10
        num_batches = num_examples // batch_size
        user_dataset = FakeUserData(gen_batch, num_batches, batch_size)
        self.user_data_test_util(user_dataset, num_examples, batch_size, num_batches)

    def test_fake_data_provider(self):
        def gen_batch(n, value=None):
            return {"data": [torch.ones(n, 10)], "label": [1] * n}

        num_batches = 2
        batch_size = 10
        num_users = 100
        fl_data_provider = FakeDataProvider(
            gen_batch, num_batches, batch_size, num_users
        )

        self.assertEqual(fl_data_provider.num_users(), num_users)
        self.assertEqual(fl_data_provider.user_ids(), list(range(num_users)))
        ad_hoc_users = [0, 3, 10, 50, 99]
        num_examples = num_batches * batch_size
        for user in ad_hoc_users:
            user_dataset = fl_data_provider.get_user_data(user)
            self.user_data_test_util(
                user_dataset, num_examples, batch_size, num_batches
            )

        self.user_data_test_util(
            fl_data_provider.test_data(), num_examples, batch_size, num_batches
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from copy import deepcopy
from typing import Type

import pytest
from flsim.channels.base_channel import FLChannelConfig, IdentityChannel
from flsim.channels.communication_stats import (
    ChannelDirection,
)
from flsim.channels.message import Message
from flsim.common.pytest_helper import assertEqual, assertIsInstance
from flsim.tests import utils
from flsim.utils.fl.common import FLModelParamUtils
from hydra.utils import instantiate


class TestIdentityChannel:
    @pytest.mark.parametrize(
        "config",
        [FLChannelConfig()],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [IdentityChannel],
    )
    def test_identity_instantiation(self, config: Type, expected_type: Type) -> None:
        """
        Tests instantiation of the identity channel.
        """

        # test instantiation
        channel = instantiate(config)
        assertIsInstance(channel, expected_type)

    @pytest.mark.parametrize(
        "config",
        [FLChannelConfig()],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [IdentityChannel],
    )
    def test_identity_server_to_client(self, config: Type, expected_type: Type) -> None:
        """
        Tests server to client transmission of the message. Models
        before and after transmission should be identical.
        """

        # instantiate channel
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        download_model = deepcopy(base_model)

        # test server -> client, models should be strictly identical
        message = Message(download_model)
        message = channel.server_to_client(message)

        mismatched = FLModelParamUtils.get_mismatched_param(
            [base_model.fl_get_module(), download_model.fl_get_module()]
        )
        assertEqual(mismatched, "", mismatched)

    @pytest.mark.parametrize(
        "config",
        [FLChannelConfig()],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [IdentityChannel],
    )
    def test_identity_client_to_server(self, config: Type, expected_type: Type) -> None:
        """
        Tests client to server transmission of the message. Models
        before and after transmission should be identical.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # test client -> server, models should be strictly identical
        message = Message(upload_model)
        message = channel.client_to_server(message)
        mismatched = FLModelParamUtils.get_mismatched_param(
            [base_model.fl_get_module(), upload_model.fl_get_module()]
        )
        assertEqual(mismatched, "", mismatched)

    @pytest.mark.parametrize(
        "config",
        [FLChannelConfig(report_communication_metrics=True)],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [IdentityChannel],
    )
    def test_identity_stats(self, config: Type, expected_type: Type) -> None:
        """
        Tests stats measurement. Bytes sent from client to server
        and from server to client should be identical.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        download_model = deepcopy(base_model)
        upload_model = deepcopy(base_model)

        # server -> client
        message = Message(download_model)
        message = channel.server_to_client(message)

        # client -> server
        message = Message(upload_model)
        message = channel.client_to_server(message)

        # test communication stats measurements
        stats = channel.stats_collector.get_channel_stats()
        client_to_server_bytes = stats[ChannelDirection.CLIENT_TO_SERVER].mean()
        server_to_client_bytes = stats[ChannelDirection.SERVER_TO_CLIENT].mean()
        assertEqual(client_to_server_bytes, server_to_client_bytes)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from copy import deepcopy
from typing import Type

import pytest
from flsim.channels.communication_stats import (
    ChannelDirection,
)
from flsim.channels.message import Message
from flsim.channels.scalar_quantization_channel import (
    ScalarQuantizationChannelConfig,
    ScalarQuantizationChannel,
)
from flsim.common.pytest_helper import assertEqual, assertIsInstance
from flsim.tests import utils
from flsim.utils.fl.common import FLModelParamUtils
from hydra.utils import instantiate


class TestScalarQuantizationChannel:
    @pytest.mark.parametrize(
        "config",
        [
            ScalarQuantizationChannelConfig(
                n_bits=8,
                quantize_per_tensor=True,
            ),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ScalarQuantizationChannel],
    )
    def test_int8_instantiation(self, config: Type, expected_type: Type) -> None:
        """
        Tests instantiation of the scalar quantization channel.
        """

        # test instantiation
        channel = instantiate(config)
        assertIsInstance(channel, expected_type)

    @pytest.mark.parametrize(
        "config",
        [
            ScalarQuantizationChannelConfig(n_bits=8, quantize_per_tensor=True),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ScalarQuantizationChannel],
    )
    def test_int8_server_to_client(self, config: Type, expected_type: Type) -> None:
        """
        Tests server to client transmission of the message. Models
        before and after transmission should be identical since we
        only scalar quantize in the client to server direction.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        download_model = deepcopy(base_model)

        # test server -> client, models should be strictly identical
        message = Message(download_model)
        message = channel.server_to_client(message)

        mismatched = FLModelParamUtils.get_mismatched_param(
            [base_model.fl_get_module(), download_model.fl_get_module()]
        )
        assertEqual(mismatched, "", mismatched)

    @pytest.mark.parametrize(
        "config",
        [
            ScalarQuantizationChannelConfig(n_bits=8, quantize_per_tensor=True),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ScalarQuantizationChannel],
    )
    def test_int8_client_to_server(self, config: Type, expected_type: Type) -> None:
        """
        Tests client to server transmission of the message. Models
        before and after transmission should be almost identical
        due to int8 emulation (n_bits = 8 here).
        """

        # instantiation
        channel = instantiate(config)
        assertIsInstance(channel, expected_type)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # test client -> server, models should be almost equal due to int8 emulation
        message = message = Message(upload_model)
        message = channel.client_to_server(message)

        mismatched = FLModelParamUtils.get_mismatched_param(
            [base_model.fl_get_module(), upload_model.fl_get_module()],
            rel_epsilon=1e-8,
            abs_epsilon=1e-2,
        )
        assertEqual(mismatched, "", mismatched)

    @pytest.mark.parametrize(
        "config",
        [
            ScalarQuantizationChannelConfig(
                n_bits=8, quantize_per_tensor=True, report_communication_metrics=True
            ),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ScalarQuantizationChannel],
    )
    def test_int8_stats(self, config: Type, expected_type: Type) -> None:
        """
        Tests stats measurement. We manually compute bytes sent from client
        to server for the two weight matrices (remember that we do not quantize
        the biases), and check that it's a quarter of the bytes sent from
        server to client (subtracting scales and zero_points).
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        download_model = deepcopy(base_model)
        upload_model = deepcopy(base_model)

        # server -> client
        message = Message(download_model)
        message = channel.server_to_client(message)

        # client -> server
        message = Message(upload_model)
        message = channel.client_to_server(message)

        # test communication stats measurements (here per_tensor int8 quantization)
        stats = channel.stats_collector.get_channel_stats()
        client_to_server_bytes = stats[ChannelDirection.CLIENT_TO_SERVER].mean()
        server_to_client_bytes = stats[ChannelDirection.SERVER_TO_CLIENT].mean()

        channels_first_layer = 5
        channels_second_layer = 1
        n_tot_channels = channels_first_layer + channels_second_layer
        bias_size_bytes = n_tot_channels * channel.BYTES_PER_FP32

        n_layers = 2
        scale_zero_point_bytes = n_layers * (
            channel.BYTES_PER_FP64 + channel.BYTES_PER_FP32
        )
        int8_weight_bytes = (
            client_to_server_bytes - bias_size_bytes - scale_zero_point_bytes
        )
        float_weights_bytes = server_to_client_bytes - bias_size_bytes
        assertEqual(4 * int8_weight_bytes, float_weights_bytes)

    @pytest.mark.parametrize(
        "config",
        [
            ScalarQuantizationChannelConfig(n_bits=4, quantize_per_tensor=False),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ScalarQuantizationChannel],
    )
    def test_quantize_constant(self, config: Type, expected_type: Type) -> None:
        """
        Tests client to server transmission of the message. Models
        before and after transmission should be identical since we
        quantize a model filled with weights and biases equal to 1.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model with constant weights
        two_fc = utils.TwoFC()
        two_fc.fill_all(1.0)
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # test client -> server, all weights are equal so quantization error is
        # zero or every acceptable value of n_bits (here tested with n_bits = 4)
        message = Message(upload_model)
        message = channel.client_to_server(message)

        mismatched = FLModelParamUtils.get_mismatched_param(
            [base_model.fl_get_module(), upload_model.fl_get_module()]
        )
        assertEqual(mismatched, "", mismatched)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from enum import Enum

from flsim.utils.fl.stats import RandomVariableStatsTrackerMA, AverageType


class TargetMetricDirection(Enum):
    MIN = "min"
    MAX = "max"


class TargetMetricTracker:
    """
    Tracks the sliding window of eval metric throughout the course of training
    and reports the round if eval metric target is reached
    """

    def __init__(
        self,
        target_value: float,
        window_size: int,
        average_type: AverageType,
        direction: TargetMetricDirection,
    ):
        self._stats = RandomVariableStatsTrackerMA(
            window_size=window_size, mode=average_type
        )
        self.target_value = target_value
        self.window_size = window_size
        self.direction = direction

    def update_and_check_target(
        self,
        current_eval_metric: float,
    ) -> bool:
        """
        Updates the stats tracker with latest eval metric
        Return value:
            True if target metric is reached.
        """
        self._stats.update(current_eval_metric)
        if self._stats.num_samples < self.window_size:
            return False
        return (
            self._stats.mean() > self.target_value
            if self.direction == TargetMetricDirection.MAX
            else self._stats.mean() < self.target_value
        )

    @property
    def mean(self):
        return self._stats.mean()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math
from collections import Counter

import torch
from flsim.active_user_selectors.simple_user_selector import (
    ActiveUserSelectorUtils,
    HighLossActiveUserSelector,
    HighLossActiveUserSelectorConfig,
    NumberOfSamplesActiveUserSelector,
    NumberOfSamplesActiveUserSelectorConfig,
    RandomRoundRobinActiveUserSelector,
    RandomRoundRobinActiveUserSelectorConfig,
    SequentialActiveUserSelector,
    SequentialActiveUserSelectorConfig,
    UniformlyRandomActiveUserSelector,
    UniformlyRandomActiveUserSelectorConfig,
)
from flsim.common.pytest_helper import (
    assertEqual,
    assertTrue,
    assertRaises,
    assertIsInstance,
)
from flsim.utils.sample_model import DummyAlphabetFLModel
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset
from hydra.utils import instantiate


class TestActiveUserSelector:
    def test_uniformly_random_user_selection(self):
        # Test the uniformly random user selection in the ActiveUserSelector base class
        null_selector = instantiate(SequentialActiveUserSelectorConfig())
        assertIsInstance(null_selector, SequentialActiveUserSelector)
        users = tuple(range(5))
        torch.manual_seed(123456789)
        selections = [
            null_selector.get_users_unif_rand(
                num_total_users=len(users), users_per_round=1
            )[0]
            for _ in range(10000)
        ]
        counts = Counter(selections)
        assertTrue(min(counts.values()) > 1500 and max(counts.values()) < 2500)
        # Make sure that the users aren't being selected sequentially
        assertTrue(min(counts.values()) < max(counts.values()) - 2)
        with assertRaises(AssertionError):
            null_selector.get_user_indices()

        uniform_selector_1 = instantiate(
            SequentialActiveUserSelectorConfig(user_selector_seed=12345)
        )
        uniform_selector_2 = instantiate(
            SequentialActiveUserSelectorConfig(user_selector_seed=12345)
        )
        selection_1 = uniform_selector_1.get_users_unif_rand(
            num_total_users=10000, users_per_round=1
        )
        selection_2 = uniform_selector_2.get_users_unif_rand(
            num_total_users=10000, users_per_round=1
        )
        assertEqual(selection_1, selection_2)

    def test_uniformly_random_user_selector(self):
        null_selector = instantiate(UniformlyRandomActiveUserSelectorConfig())
        assertIsInstance(null_selector, UniformlyRandomActiveUserSelector)
        users = tuple(range(5))
        torch.manual_seed(123456789)
        selections = [
            null_selector.get_user_indices(
                num_total_users=len(users), users_per_round=1
            )[0]
            for _ in range(10000)
        ]
        counts = Counter(selections)
        assertTrue(min(counts.values()) > 1500 and max(counts.values()) < 2500)
        with assertRaises(AssertionError):
            null_selector.get_user_indices()

        uniform_selector_1 = instantiate(
            UniformlyRandomActiveUserSelectorConfig(user_selector_seed=12345)
        )
        uniform_selector_2 = instantiate(
            UniformlyRandomActiveUserSelectorConfig(user_selector_seed=12345)
        )
        selection_1 = uniform_selector_1.get_user_indices(
            num_total_users=10000, users_per_round=1
        )
        selection_2 = uniform_selector_2.get_user_indices(
            num_total_users=10000, users_per_round=1
        )
        assertEqual(selection_1, selection_2)

    def test_sequential_user_selector(self):
        # 1) test if num_users is not divisible by users_per_round
        num_total_users, users_per_round, round_num = 10, 3, 8
        selector = instantiate(SequentialActiveUserSelectorConfig())
        assertIsInstance(selector, SequentialActiveUserSelector)
        for round_index in range(round_num):
            user_indices = selector.get_user_indices(
                num_total_users=num_total_users, users_per_round=users_per_round
            )
            if round_index % 4 == 0:
                assertEqual(user_indices, [0, 1, 2])
            if round_index % 4 == 1:
                assertEqual(user_indices, [3, 4, 5])
            if round_index % 4 == 2:
                assertEqual(user_indices, [6, 7, 8])
            if round_index % 4 == 3:
                assertEqual(user_indices, [9])

        # 2) test if num_users is divisible by users_per_round
        num_total_users, round_num = 9, 6
        selector = instantiate(SequentialActiveUserSelectorConfig())
        for round_index in range(round_num):
            user_indices = selector.get_user_indices(
                num_total_users=num_total_users, users_per_round=users_per_round
            )
            if round_index % 3 == 0:
                assertEqual(user_indices, [0, 1, 2])
            if round_index % 3 == 1:
                assertEqual(user_indices, [3, 4, 5])
            if round_index % 3 == 2:
                assertEqual(user_indices, [6, 7, 8])

    def test_random_round_robin_user_selector(self):
        # 1) test if num_users is not divisible by users_per_round
        num_total_users, users_per_round = 87, 15
        available_users = range(num_total_users)
        selector = instantiate(RandomRoundRobinActiveUserSelectorConfig())
        assertIsInstance(selector, RandomRoundRobinActiveUserSelector)
        while len(available_users) > 0:
            user_indices = selector.get_user_indices(
                num_total_users=num_total_users, users_per_round=users_per_round
            )
            user_indices_set = set(user_indices)
            available_user_set = set(available_users)
            assertEqual(len(user_indices), min(users_per_round, len(available_users)))
            assertEqual(len(user_indices_set), len(user_indices))
            assertEqual(len(available_user_set), len(available_users))
            assertTrue(user_indices_set.issubset(available_user_set))
            available_users = list(available_user_set - user_indices_set)

        # Try one more time to make sure that the user selector class reset
        user_indices = selector.get_user_indices(
            num_total_users=num_total_users, users_per_round=users_per_round
        )
        user_indices_set = set(user_indices)
        available_user_set = set(range(num_total_users))
        assertEqual(len(user_indices), users_per_round)
        assertEqual(len(user_indices_set), users_per_round)
        assertTrue(user_indices_set.issubset(available_user_set))

        # 2) test if num_users is divisible by users_per_round
        num_total_users, users_per_round = 60, 15
        selector = instantiate(RandomRoundRobinActiveUserSelectorConfig())
        while len(available_users) > 0:
            user_indices = selector.get_user_indices(
                num_total_users=num_total_users, users_per_round=users_per_round
            )
            user_indices_set = set(user_indices)
            available_user_set = set(available_users)
            assertEqual(len(user_indices), users_per_round)
            assertEqual(len(user_indices_set), len(user_indices))
            assertEqual(len(available_user_set), len(available_users))
            assertTrue(user_indices_set.issubset(available_user_set))
            available_users = list(available_user_set - user_indices_set)

        # Try one more time to make sure that the user selector class reset
        user_indices = selector.get_user_indices(
            num_total_users=num_total_users, users_per_round=users_per_round
        )
        user_indices_set = set(user_indices)
        available_user_set = set(range(num_total_users))
        assertEqual(len(user_indices), users_per_round)
        assertEqual(len(user_indices_set), users_per_round)
        assertTrue(user_indices_set.issubset(available_user_set))

    def test_number_of_samples_user_selector(self):
        selector = instantiate(NumberOfSamplesActiveUserSelectorConfig())
        assertIsInstance(selector, NumberOfSamplesActiveUserSelector)

        shard_size = 5
        local_batch_size = 4
        dummy_dataset = DummyAlphabetDataset(num_rows=6)
        dummy_model = DummyAlphabetFLModel()
        data_provider, _ = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, dummy_model
        )

        selections = [
            selector.get_user_indices(users_per_round=1, data_provider=data_provider)[0]
            for _ in range(1000)
        ]
        counts = Counter(selections)
        assertTrue(counts[0] > counts[1])

    def test_high_loss_user_selector(self):
        selector = instantiate(HighLossActiveUserSelectorConfig())
        assertIsInstance(selector, HighLossActiveUserSelector)

        selector = instantiate(
            HighLossActiveUserSelectorConfig(softmax_temperature=0.01)
        )

        shard_size = 10
        local_batch_size = 4
        dummy_dataset = DummyAlphabetDataset(num_rows=11)

        dummy_model = DummyAlphabetFLModel()
        data_provider, _ = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, dummy_model
        )

        selections = [
            selector.get_user_indices(
                num_total_users=2,
                users_per_round=1,
                data_provider=data_provider,
                global_model=dummy_model,
                epoch=0,
            )[0]
            for i in range(1000)
        ]
        counts = Counter(selections)
        assertTrue(counts[0] > counts[1])

        selector = instantiate(
            HighLossActiveUserSelectorConfig(
                softmax_temperature=0.01, epochs_before_active=1
            )
        )

        selections = [
            selector.get_user_indices(
                num_total_users=2,
                users_per_round=1,
                data_provider=data_provider,
                global_model=dummy_model,
                epoch=0,
            )[0]
            for i in range(1000)
        ]
        counts = Counter(selections)
        assertTrue(counts[0] > 400)
        assertTrue(counts[0] < 600)


class TestActiveUserSelectorUtils:

    tolerence = 1e-5

    def test_convert_to_probability(self):
        valuations = torch.tensor([1, 1, 1, 2, 2], dtype=torch.float)
        weights = torch.tensor([1, 1, 1, 1, 1], dtype=torch.float)

        total_prob = sum(
            ActiveUserSelectorUtils.convert_to_probability(valuations, 0, 1)
        )

        assertEqual(total_prob, 1)

        prob_1 = ActiveUserSelectorUtils.convert_to_probability(
            valuations, 0, 1, weights
        )
        prob_2 = ActiveUserSelectorUtils.convert_to_probability(valuations, 0, 1)
        assertTrue(torch.allclose(prob_1, prob_2, rtol=self.tolerence))

        prob_1 = ActiveUserSelectorUtils.convert_to_probability(valuations, 0, 1)
        prob_2 = torch.exp(valuations) / sum(torch.exp(valuations))
        assertTrue(torch.allclose(prob_1, prob_2, rtol=self.tolerence))

        weights = torch.tensor([1, 2, 1, 2, 1], dtype=torch.float)
        unnormalized_probs = torch.tensor(
            [math.exp(1), 2 * math.exp(1), math.exp(1), 2 * math.exp(2), math.exp(2)],
            dtype=torch.float,
        )
        prob_1 = unnormalized_probs / sum(unnormalized_probs)
        prob_2 = ActiveUserSelectorUtils.convert_to_probability(
            valuations, 0, 1, weights
        )
        assertTrue(torch.allclose(prob_1, prob_2, rtol=self.tolerence))

        prob_1 = ActiveUserSelectorUtils.convert_to_probability(valuations, 0, 0)
        prob_2 = torch.tensor([0.2] * 5, dtype=torch.float)
        assertTrue(torch.allclose(prob_1, prob_2, rtol=self.tolerence))

        prob_1 = ActiveUserSelectorUtils.convert_to_probability(valuations, 0, 25)
        prob_2 = torch.tensor([0, 0, 0, 0.5, 0.5], dtype=torch.float)
        assertTrue(torch.allclose(prob_1, prob_2, rtol=self.tolerence))

        prob = ActiveUserSelectorUtils.convert_to_probability(valuations, 0.5, 1)
        assertEqual(len(torch.nonzero(prob)), 3)

        with assertRaises(AssertionError):
            ActiveUserSelectorUtils.convert_to_probability(valuations, 1, 1)

    def test_normalize_by_sample_count(self):
        user_utility = torch.tensor([0, 1, 2, 3], dtype=torch.float)
        counts = torch.tensor([1, 10, 100, 1], dtype=torch.float)

        no_normalization = ActiveUserSelectorUtils.normalize_by_sample_count(
            user_utility, counts, 0
        )
        assertTrue(torch.allclose(no_normalization, user_utility, rtol=self.tolerence))

        avged = user_utility / counts
        avg_normalization = ActiveUserSelectorUtils.normalize_by_sample_count(
            user_utility, counts, 1
        )
        assertTrue(torch.allclose(avg_normalization, avged, rtol=self.tolerence))

    def test_samples_per_user(self):
        shard_size = 4
        local_batch_size = 4
        dummy_dataset = DummyAlphabetDataset()

        dummy_model = DummyAlphabetFLModel()
        data_provider, _ = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, dummy_model
        )
        samples_per_user = ActiveUserSelectorUtils.samples_per_user(data_provider)
        assertTrue(
            torch.allclose(
                samples_per_user, torch.tensor([4, 4, 4, 4, 4, 4, 2], dtype=torch.float)
            )
        )

    def test_select_users(self):
        """select_users has two mechanisms for selecting users: p proportion are
        selected according to the probabilties in probs, and (1-p) proportion are
        selected uniformly at random, where p = fraction_uniformly_random. This test
        ensures that the correct number of users are returned for different values
        of p.
        """
        probs = torch.tensor([0.1] * 10, dtype=torch.float)
        users_per_round = 3
        rng = torch.Generator()

        assertEqual(
            len(ActiveUserSelectorUtils.select_users(users_per_round, probs, 0, rng)), 3
        )
        assertEqual(
            len(ActiveUserSelectorUtils.select_users(users_per_round, probs, 1, rng)), 3
        )
        assertEqual(
            len(ActiveUserSelectorUtils.select_users(users_per_round, probs, 0.5, rng)),
            3,
        )
        assertEqual(
            sorted(
                ActiveUserSelectorUtils.select_users(
                    2, torch.tensor([0.5, 0.5, 0, 0, 0], dtype=torch.float), 0, rng
                )
            ),
            [0, 1],
        )

    def test_sample_available_users(self):
        num_total_users, users_per_round = 95, 10
        available_users = range(num_total_users)
        rng = torch.Generator()

        while len(available_users) > 0:
            prev_available_users = available_users
            user_indices = ActiveUserSelectorUtils.sample_available_users(
                users_per_round, available_users, rng
            )
            user_indices_set = set(user_indices)
            available_user_set = set(available_users)

            assertEqual(len(user_indices), min(users_per_round, len(available_users)))
            assertEqual(available_users, prev_available_users)
            assertTrue(user_indices_set.issubset(available_user_set))
            assertEqual(len(user_indices_set), len(user_indices))
            assertEqual(len(available_user_set), len(available_users))

            available_users = list(available_user_set - user_indices_set)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
from typing import Any, List

import torch


class IFLBatchMetrics(abc.ABC):
    """Each forward run of FL Model (i.e. concrete implementation of IFLModel
    in PyML) will return an IFLBatchMetrics object. This is an encapsulation
    and abstraction around several useful/common metrics such as loss,
    num_examples, predictions, and targets so that IFLMetricsReporter can
    aggregate metrics from any kind of model as long as the model returns
    IFLBatchMetrics.
    """

    @property
    @abc.abstractmethod
    def loss(self) -> torch.Tensor:
        pass

    @property
    @abc.abstractmethod
    def num_examples(self) -> int:
        pass

    # TODO: ngjhn Remove these since they accumulate in memory causing OOM
    @property
    @abc.abstractmethod
    def predictions(self) -> List[Any]:
        pass

    @property
    @abc.abstractmethod
    def targets(self) -> List[Any]:
        pass

    @property
    @abc.abstractmethod
    def model_inputs(self) -> Any:
        pass



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math
import random
import string

import torch
from flsim.common.pytest_helper import assertEqual, assertTrue
from flsim.data.data_sharder import (
    RandomSharderConfig,
    SequentialSharderConfig,
    RoundRobinSharderConfig,
    ColumnSharderConfig,
    BroadcastSharderConfig,
    PowerLawSharderConfig,
)
from flsim.data.dataset_data_loader import FLDatasetDataLoaderWithBatch
from flsim.utils.sample_model import TestDataSetting
from flsim.utils.tests.helpers.test_data_utils import (
    DummyAlphabetDataset,
    Utils,
)
from hydra.utils import instantiate


class MockData:
    @staticmethod
    def provide_data():
        """Generate 26 rows of data. Each row:
        "label":0 or 1, "text":int between 0 and 25
        "user_id": between 'a' and 'z'
        """
        characters = Utils.get_characters(num=26)
        return [
            {
                TestDataSetting.LABEL_COL_NAME: Utils.get_label(character),
                TestDataSetting.TEXT_COL_NAME: Utils.get_text(character),
                TestDataSetting.USER_ID_COL_NAME: torch.tensor(
                    ord(character), dtype=torch.int8
                ),
            }
            for character in characters
        ]


class TestDataSharder:
    def test_random_sharder(self) -> None:
        """Only tests random sharding strategy in this test case. All the other
        strategies are tested in test_shard_rows().
        """
        random.seed(1)
        random_sharder = instantiate(
            RandomSharderConfig(num_shards=TestDataSetting.NUM_SHARDS)
        )
        for i in range(random_sharder.cfg.num_shards + 1):
            shard = random_sharder.shard_for_row(
                MockData.provide_data()[i % sum(1 for row in MockData.provide_data())]
            )
            if i == 0:
                assertEqual(shard, [6])
            elif i == random_sharder.cfg.num_shards:
                assertEqual(shard, [3])

    def test_shard_rows_random(self) -> None:
        random_sharder = instantiate(
            RandomSharderConfig(num_shards=TestDataSetting.NUM_SHARDS)
        )
        random_sharded_rows = list(
            random_sharder.shard_rows(
                MockData.provide_data()[
                    : min(
                        int(random_sharder.cfg.num_shards * 1.2 + 1),
                        sum(1 for row in MockData.provide_data()),
                    )
                ]
            )
        )
        assertEqual(len(random_sharded_rows), random_sharder.cfg.num_shards)

    def test_shard_rows_power_law(self) -> None:
        power_law_sharder = instantiate(PowerLawSharderConfig(num_shards=2, alpha=0.2))
        power_law_sharded_rows = list(
            power_law_sharder.shard_rows(MockData.provide_data())
        )
        num_examples_per_user = [len(p) for _, p in power_law_sharded_rows]
        assertEqual(len(num_examples_per_user), power_law_sharder.cfg.num_shards)
        # assert that top user will have more than 20% of the examples
        assertTrue(max(num_examples_per_user) / sum(num_examples_per_user) > 0.20)

    def test_shard_rows_broadcast(self) -> None:
        broadcast_sharder = instantiate(
            BroadcastSharderConfig(num_shards=TestDataSetting.NUM_SHARDS)
        )
        # all rows from dataset should be replicated to all shards
        assertEqual(
            [
                user_data
                for _, user_data in broadcast_sharder.shard_rows(
                    MockData.provide_data()
                )
            ],
            [
                MockData.provide_data()
                for shard_idx in range(TestDataSetting.NUM_SHARDS)
            ],
        )

    def test_shard_rows_column(self) -> None:
        column_sharder = instantiate(
            ColumnSharderConfig(sharding_col=TestDataSetting.USER_ID_COL_NAME)
        )
        # each data row should be assigned to a unique shard, since column
        # sharding is based on user id here and each mocked user id is unique
        assertEqual(
            [
                user_data
                for _, user_data in column_sharder.shard_rows(MockData.provide_data())
            ],
            [[one_data_row] for one_data_row in MockData.provide_data()],
        )

    def test_shard_rows_round_robin(self) -> None:
        round_robin_sharder = instantiate(
            RoundRobinSharderConfig(num_shards=TestDataSetting.NUM_SHARDS)
        )
        sharded_rows_from_round_robin = [
            user_data
            for _, user_data in round_robin_sharder.shard_rows(MockData.provide_data())
        ]
        # there are 26 data rows here and it should be sharded with round-
        # robin fashion with 10 shards. e.g. 1th, 11th, 21th data row
        # should be in the same shard.
        for shard_index in range(round_robin_sharder.cfg.num_shards):
            assertEqual(
                sharded_rows_from_round_robin[shard_index],
                [
                    one_data_row
                    for row_index, one_data_row in enumerate(MockData.provide_data())
                    if row_index % round_robin_sharder.cfg.num_shards == shard_index
                ],
            )

    def test_shard_rows_sequential(self) -> None:
        sequential_sharder = instantiate(SequentialSharderConfig(examples_per_shard=2))
        sharded_rows_from_sequential = [
            user_data
            for _, user_data in sequential_sharder.shard_rows(MockData.provide_data())
        ]
        assertEqual(
            len(sharded_rows_from_sequential),
            len(string.ascii_lowercase) / sequential_sharder.cfg.examples_per_shard,
        )
        dataset = MockData.provide_data()
        for shard_index, data_for_a_shard in enumerate(sharded_rows_from_sequential):
            start_index = shard_index * sequential_sharder.cfg.examples_per_shard
            assertEqual(
                data_for_a_shard,
                dataset[
                    start_index : start_index
                    + sequential_sharder.cfg.examples_per_shard
                ],
            )

    def test_distributed_user_sharding(self) -> None:
        # mock 26 rows
        shard_size = 4
        local_batch_size = 2
        world_size = 4

        # mock world_size parallel creation of fl_train_set
        for rank in range(world_size):
            fl_data_sharder = instantiate(
                SequentialSharderConfig(examples_per_shard=shard_size)
            )
            data_loader = FLDatasetDataLoaderWithBatch(
                MockData.provide_data(),
                MockData.provide_data(),
                MockData.provide_data(),
                fl_data_sharder,
                local_batch_size,  # train_batch_size
                local_batch_size,  # eval_batch_size
                local_batch_size,  # test_batch_size
            )
            train_set = data_loader.fl_train_set(rank=rank, world_size=world_size)
            assertEqual(data_loader.num_total_users, math.ceil(26 / shard_size))
            number_of_users_on_worker = 2 if rank in (0, 1, 2) else 1
            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Iterable[typing.Iterable[typing.Any]]`.
            train_set_size = len(train_set)
            assertEqual(train_set_size, number_of_users_on_worker)

    def test_pytorch_dataset_wrapper(self) -> None:
        # mock 26 rows
        shard_size = 4
        local_batch_size = 2

        fl_data_sharder = instantiate(
            SequentialSharderConfig(examples_per_shard=shard_size)
        )
        dummy_dataset = DummyAlphabetDataset()
        data_loader = FLDatasetDataLoaderWithBatch(
            dummy_dataset,
            dummy_dataset,
            dummy_dataset,
            fl_data_sharder,
            local_batch_size,  # train_batch_size
            local_batch_size,  # eval_batch_size
            local_batch_size,  # test_batch_size
        )

        data_loader.fl_train_set()
        assertEqual(data_loader.num_total_users, math.ceil(26 / shard_size))
        eval_set = data_loader.fl_eval_set()
        # pyre-fixme[6]
        assertEqual(len(eval_set), (26 / local_batch_size))



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore  #  @manual

from .timeout_simulator import (
    NeverTimeOutSimulatorConfig,
    GaussianTimeOutSimulatorConfig,
)


ConfigStore.instance().store(
    name="base_never_timeout_simulator",
    node=NeverTimeOutSimulatorConfig,
    group="timeout_simulator",
)


ConfigStore.instance().store(
    name="base_gaussian_timeout_simulator",
    node=GaussianTimeOutSimulatorConfig,
    group="timeout_simulator",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional

import torch
from flsim.channels.message import Message
from flsim.channels.sketch_channel import SketchChannel
from flsim.common.logger import Logger
from flsim.interfaces.model import IFLModel
from flsim.reducers.base_round_reducer import (
    IFLRoundReducerConfig,
    IFLRoundReducer,
    ReductionType,
    ReductionPrecision,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.count_sketch import CountSketch


class SketchRoundReducer(IFLRoundReducer):

    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: SketchChannel,
        num_users_per_round: Optional[int] = None,
        total_number_of_users: Optional[int] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        if channel is None or not isinstance(channel, SketchChannel):
            raise AssertionError(f"Expected a SketchChannel, got {channel}")

        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=SketchRoundReducerConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

        device = next(global_model.fl_get_module().parameters()).device
        self.round_cs = CountSketch(
            # pyre-fixme[16]: `SketchChannel` has no attribute `cfg`.
            width=channel.cfg.num_col,
            depth=channel.cfg.num_hash,
            prime=channel.cfg.prime,
            independence=channel.cfg.independence,
            h=channel.h,
            g=channel.g,
            device=device,
        )
        self.channel = channel

        self.num_users_per_round = num_users_per_round
        self.total_number_of_users = total_number_of_users

        self.h = channel.h
        self.g = channel.g

        # pyre-fixme[16]: `SketchRoundReducer` has no attribute `cfg`.
        self.reduction_type = self.cfg.reduction_type
        self.dtype = self.cfg.precision.dtype
        self.sum_weights: torch.Tensor = torch.zeros(1, device=device)
        self.reset(global_model)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def set_num_total_users(self, num_total_users):
        self.total_number_of_users = num_total_users

    def reset(self, ref_model: IFLModel) -> None:
        """
        Initializes / Resets round reducers internals given a model.
        """
        self.round_cs.set_params(ref_model.fl_get_module().state_dict())
        self.sum_weights.fill_(0)

    def collect_update(self, delta: IFLModel, weight: float) -> None:

        # build channel message with model state dict
        # receive through channel
        message = self.channel.client_to_server(
            Message(model=delta, weight=weight, count_sketch=CountSketch())
        )

        # decode channel message, here only get the sketch
        sketch = message.count_sketch

        if not self.is_weighted:
            weight = 1.0

        self.round_cs.linear_comb(1.0, sketch, weight)
        self.sum_weights += weight

    def reduce(self):
        if torch.isclose(
            self.sum_weights, torch.tensor([0.0], device=self.sum_weights.device)
        ):
            raise AssertionError(
                "Cannot call reduce when no updates have been collected."
            )
        self.round_cs.buckets /= self.sum_weights
        return self.round_cs, self.sum_weights

    @property
    def is_weighted(self):
        return self.reduction_type in (
            ReductionType.WEIGHTED_SUM,
            ReductionType.WEIGHTED_AVERAGE,
        )


@dataclass
class SketchRoundReducerConfig(IFLRoundReducerConfig):
    _target_: str = fullclassname(SketchRoundReducer)
    only_federated_params: bool = False
    # TODO: Handle reduction_type and precision T95227332
    reduction_type: ReductionType = ReductionType.WEIGHTED_AVERAGE
    precision: ReductionPrecision = ReductionPrecision.DEFAULT
